from typing import Any

import numpy as np
import pandas as pd
from anndata import AnnData


def get_obs_indices_for_cluster(adata: AnnData, obs_column: str, value: Any) -> np.ndarray:
    """
    Get the indices of observations in a specific cluster.

    :param adata: Annotated data object.
    :type adata: AnnData
    :param obs_column: Name of the observation column to filter on.
    :type obs_column: str
    :param value: Value to filter the observation column on.
    :type value: Any
    :return: Array of indices of observations in the specified cluster.
    :rtype: np.ndarray
    """
    assert obs_column in adata.obs.columns
    assert isinstance(adata.obs[obs_column].dtype, pd.CategoricalDtype)
    assert value in adata.obs[obs_column].cat.categories
    return (adata.obs[obs_column] == value).values.nonzero()[0]



import os
import re
from sys import version_info as python_version_info


def get_version() -> str:
    """
    Get the version of the application based on the git tag.

    :return: The version of the application.
    """

    if python_version_info.major == 3 and python_version_info.minor >= 8:
        from importlib.metadata import PackageNotFoundError, version

        try:
            return version("cellarium-cas")
        except PackageNotFoundError:
            return os.environ.get("CAS_VERSION", "0.0.1")
    elif python_version_info.major == 3 and python_version_info.minor == 7:
        from importlib_metadata import PackageNotFoundError, version

        try:
            return version("cellarium-cas")
        except PackageNotFoundError:
            return os.environ.get("CAS_VERSION", "0.0.1")
    else:
        raise Exception("Unsupported Python version. Please use Python 3.7 or later.")


def _is_release_version(version_string) -> bool:
    """
    Check if the version string is a release version. Release version is a string that has the format of
    "major.minor.patch".

    Reference: https://semver.org/

    :param version_string: The version string to check

    :return: True if the version string is a release version, False otherwise.
    """
    pattern = r"^\d+\.\d+\.\d+$"
    return re.match(pattern, version_string) is not None


def _is_pre_release_version(version_string) -> bool:
    """
    Check if the version string is a pre-release version. Pre-release version is a string that has the format of
    "major.minor.patch-(alpha|beta|rc).patch".

    Reference: https://semver.org/

    :param version_string: The version string to check

    :return: True if the version string is a pre-release version, False otherwise.
    """
    pattern = r"^\d+\.\d+\.\d+-(alpha|beta|rc)\.\d+$"
    return re.match(pattern, version_string) is not None


def get_version_environment() -> str:
    """
    Get the environment based on the version of the application. The environment is considered as "development" if the
    version is a pre-release version, "production" if the version is a release version, and "test" if the version is
    neither a release version nor a pre-release version.

    Exception: If the version is "0.0.1", the environment is considered as "development" because this version is
    assigned in GitHub Actions executed on branches without a tag.

    :return: The environment name of the application.
    """
    app_version = get_version()
    if _is_pre_release_version(app_version) or app_version == "0.0.1":
        return "development"
    elif _is_release_version(app_version):
        return "production"
    else:
        return "test"



from enum import Enum


class HTTP:
    """
    HTTP status codes constants
    """

    STATUS_200_OK = 200
    STATUS_201_CREATED = 201
    STATUS_202_ACCEPTED = 202
    STATUS_204_NO_CONTENT = 204
    STATUS_226_IM_USED = 226

    STATUS_301_MOVED_PERMANENTLY = 301
    STATUS_308_PERMANENT_REDIRECT = 308

    STATUS_400_BAD_REQUEST = 400
    STATUS_401_UNAUTHORIZED = 401
    STATUS_403_FORBIDDEN = 403
    STATUS_NOT_FOUND = 404

    STATUS_500_INTERNAL_SERVER_ERROR = 500
    STATUS_503_SERVICE_UNAVAILABLE = 503
    STATUS_504_GATEWAY_TIMEOUT = 504
    STATUS_511_NETWORK_AUTHENTICATION_REQUIRED = 511


class CountMatrixInput(Enum):
    X: str = "X"
    RAW_X: str = "raw.X"


class Headers:
    # The authorization header.
    authorization = "Authorization"
    # The client session id that is used to track a user's CAS client session.
    client_session_id = "x-client-session-id"
    # The client action id that is used to track a user's logical action that may span multiple requests.
    client_action_id = "x-client-action-id"



import asyncio
import datetime
import functools
import math
import operator
import pkgutil
import time
import typing as t
import warnings
from contextlib import contextmanager
from uuid import UUID, uuid4

import anndata
from anndata import ImplicitModificationWarning
from deprecated import deprecated

from cellarium.cas.service import action_context_manager

from . import _io, constants, exceptions, preprocessing, service, settings, version


@contextmanager
def suppress_implicit_modification_warning():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", ImplicitModificationWarning)
        yield


CHUNK_SIZE_ANNOTATE_DEFAULT = 1000
CHUNK_SIZE_SEARCH_DEFAULT = 500
DEFAULT_PRUNE_THRESHOLD = 0.05
DEFAULT_WEIGHTING_PREFACTOR = 1.0

FEEDBACK_TEMPLATE = pkgutil.get_data(__name__, "resources/feedback_template.html").decode("utf-8")


class CASClient:
    """
    Service that is designed to communicate with the Cellarium Cloud Backend.

    :param api_token: API token issued by the Cellarium team
    :param num_attempts_per_chunk: Number of attempts the client should make to annotate each chunk. |br|
        `Default:` ``3``
    """

    def _print_models(self, models):
        s = "Allowed model list in Cellarium CAS:\n"
        for model in models:
            model_name = model["model_name"]
            description = model["description"]
            model_schema = model["schema_name"]
            embedding_dimension = model["embedding_dimension"]
            if model["is_default_model"]:
                model_name += " (default)"

            s += f"  - {model_name}\n    Description: {description}\n    Schema: {model_schema}\n    Embedding dimension: {embedding_dimension}\n"

        self._print(s)

    @action_context_manager()
    def __init__(
        self,
        api_token: str,
        api_url: str = settings.CELLARIUM_CLOUD_BACKEND_URL,
        num_attempts_per_chunk: int = settings.NUM_ATTEMPTS_PER_CHUNK_DEFAULT,
    ) -> None:
        self.client_session_id: UUID = uuid4()
        self.cas_api_service = service.CASAPIService(
            api_token=api_token, api_url=api_url, client_session_id=self.client_session_id
        )

        self._print(f"Connecting to the Cellarium Cloud backend with session {self.client_session_id}...")
        self.user_info = self.cas_api_service.validate_token()
        username = self.user_info["username"]
        self._print(f"User is {username}")
        self.should_show_feedback = True
        if "should_ask_for_feedback" in self.user_info:
            self.should_show_feedback = self.user_info["should_ask_for_feedback"]

        self.validate_version()

        # Retrieving General Info
        application_info = self.cas_api_service.get_application_info()

        self.model_objects_list = self.cas_api_service.get_model_list()
        self.allowed_models_list = [x["model_name"] for x in self.model_objects_list]
        self._model_name_obj_map = {x["model_name"]: x for x in self.model_objects_list}
        self.default_model_name = [x["model_name"] for x in self.model_objects_list if x["is_default_model"]][0]

        self.feature_schemas = self.cas_api_service.get_feature_schemas()
        self._feature_schemas_cache = {}

        self.num_attempts_per_chunk = num_attempts_per_chunk
        self._print(f"Authenticated in Cellarium Cloud v. {application_info['application_version']}")

        self._print_models(self.model_objects_list)

    @staticmethod
    def _get_number_of_chunks(adata, chunk_size):
        return math.ceil(len(adata) / chunk_size)

    @staticmethod
    def _get_timestamp() -> str:
        return datetime.datetime.now().strftime("%H:%M:%S.%f")[:-3]

    def _print(self, str_to_print: str) -> None:
        print(f"* [{self._get_timestamp()}] {str_to_print}")

    def _render_feedback_link(self):
        try:
            if settings.is_interactive_environment() and self.should_show_feedback:
                # only import IPython if we are in an interactive environment
                from IPython.display import HTML, display

                display(HTML(FEEDBACK_TEMPLATE.format(link=self.cas_api_service.get_feedback_answer_link())))
        except ModuleNotFoundError:
            pass

    def feedback_opt_out(self):
        self.should_show_feedback = False
        self.user_info = self.cas_api_service.feedback_opt_out()
        self._print("Successfully opted out. You will no longer receive requests to provide feedback.")

    def validate_version(self):
        """
        Validate that this version of the client library is compatible with the selected server.
        """
        client_version = version.get_version()
        version_validation_info = self.cas_api_service.validate_version(version_str=client_version)
        if version_validation_info["is_valid"]:
            self._print(f"Client version {client_version} is compatible with selected server.")
        else:
            raise exceptions.ClientTooOldError(
                f"Client version {client_version} is older than the minimum version for this server {version_validation_info['min_version']}. "
                f"Please update the client to the latest version using 'pip install cellarium-cas --upgrade'."
            )

    def validate_and_sanitize_input_data(
        self,
        adata: anndata.AnnData,
        cas_model_name: str,
        count_matrix_name: constants.CountMatrixInput,
        feature_ids_column_name: str,
        feature_names_column_name: t.Optional[str] = None,
    ) -> anndata.AnnData:
        """
        Validate and sanitize input :class:`anndata.AnnData` instance according to a specified feature schema associated
        with a particular model.

        :param adata: :class:`anndata.AnnData` instance to annotate
        :param cas_model_name: The model associated with the schema used for sanitizing. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list keyword, which refers to the
            default selected model in the Cellarium backend. |br|
        :param count_matrix_name:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values:` Choice of either ``"X"``  or ``"raw.X"`` in order to use ``adata.X`` or ``adata.raw.X``|br|
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``
        :return: Validated and sanitized instance of :class:`anndata.AnnData`
        """
        cas_model_obj = self._model_name_obj_map[cas_model_name]
        feature_schema_name = cas_model_obj["schema_name"]
        try:
            cas_feature_schema_list = self._feature_schemas_cache[feature_schema_name]
        except KeyError:
            cas_feature_schema_list = self.cas_api_service.get_feature_schema_by(name=feature_schema_name)
            self._feature_schemas_cache[feature_schema_name] = cas_feature_schema_list

        try:
            preprocessing.validate(
                adata=adata,
                cas_feature_schema_list=cas_feature_schema_list,
                feature_ids_column_name=feature_ids_column_name,
            )
        except exceptions.DataValidationError as e:
            if e.extra_features > 0:
                self._print(
                    f"The input data matrix has {e.extra_features} extra features compared to '{feature_schema_name}' "
                    f"CAS schema ({len(cas_feature_schema_list)}). "
                    f"Extra input features will be dropped."
                )
            if e.missing_features > 0:
                self._print(
                    f"The input data matrix has {e.missing_features} missing features compared to "
                    f"'{feature_schema_name}' CAS schema ({len(cas_feature_schema_list)}). "
                    f"Missing features will be imputed with zeros."
                )
            if e.extra_features == 0 and e.missing_features == 0:
                self._print(
                    f"Input datafile has all the necessary features as {feature_schema_name}, but it's still "
                    f"incompatible because of the different order. The features will be reordered "
                    f"according to {feature_schema_name}..."
                )
                self._print(
                    f"The input data matrix contains all of the features specified in '{feature_schema_name}' "
                    f"CAS schema but in a different order. The input features will be reordered according to "
                    f"'{feature_schema_name}'"
                )
            return preprocessing.sanitize(
                adata=adata,
                cas_feature_schema_list=cas_feature_schema_list,
                count_matrix_input=count_matrix_name,
                feature_ids_column_name=feature_ids_column_name,
                feature_names_column_name=feature_names_column_name,
            )
        else:
            self._print(f"The input data matrix conforms with the '{feature_schema_name}' CAS schema.")
            return adata

    def print_user_quota(self) -> None:
        """
        Print the user's quota information
        """
        user_quota = self.cas_api_service.get_user_quota()
        self._print(
            f"User quota: {user_quota['quota']}, Remaining quota: {user_quota['remaining_quota']}, "
            f"Reset date: {user_quota['quota_reset_date']}"
        )

    def __get_async_sharded_request_callback(
        self,
        results: t.List,
        chunk_index: int,
        chunk_start_i: int,
        chunk_end_i: int,
        semaphore: asyncio.Semaphore,
        service_request_callback: t.Callable,
    ) -> t.Callable:
        """
        A wrapper around POST request that handles HTTP and client errors, and resubmits the task if necessary.

        In case of HTTP 500, 503, 504 or ClientError print a message and resubmit the task.
        In case of HTTP 401 print a message to check the token.
        In case of any other error print a message.

        :param results: Results list that needs to be used to inplace the response from the server
        :param chunk_index: Consequent number of the chunk (e.g. Chunk 1, Chunk 2)
        :param chunk_start_i: Index pointing to the main adata file start position of the current chunk
        :param chunk_end_i: Index pointing to the main adata file end position of the current chunk
        :param semaphore: Semaphore object to limit the number of concurrent requests at a time
        :param service_request_callback: Callback function to execute (should be one of the async methods of the
            :class:`CASAPIService`)

        :return: A callback function that can be used to submit a request to the backend
        """

        async def sharded_request_task(**callback_kwargs):
            async with semaphore:
                retry_delay = settings.START_RETRY_DELAY

                for _ in range(self.num_attempts_per_chunk):
                    try:
                        results[chunk_index] = await service_request_callback(**callback_kwargs)

                    except (exceptions.HTTPError5XX, exceptions.HTTPClientError) as e:
                        self._print(str(e))
                        self._print(
                            f"Resubmitting chunk #{chunk_index + 1:2.0f} ({chunk_start_i:5.0f}, "
                            f"{chunk_end_i:5.0f}) to CAS ..."
                        )
                        await asyncio.sleep(retry_delay)
                        retry_delay = min(retry_delay * 2, settings.MAX_RETRY_DELAY)
                        continue
                    except exceptions.HTTPError401:
                        self._print("Unauthorized token. Please check your API token or request a new one.")
                        break
                    except exceptions.HTTPError403 as e:
                        self._print(str(e))
                        break
                    except Exception as e:
                        self._print(f"Unexpected error: {e.__class__.__name__}; Message: {str(e)}")
                        break
                    else:
                        self._print(
                            f"Received the result for cell chunk #{chunk_index + 1:2.0f} ({chunk_start_i:5.0f}, "
                            f"{chunk_end_i:5.0f}) ..."
                        )
                        break

        return sharded_request_task

    @suppress_implicit_modification_warning()
    def __async_sharded_request(
        self,
        adata: anndata.AnnData,
        chunk_size: int,
        request_callback: t.Callable,
        request_callback_kwargs: t.Dict[str, t.Any],
    ) -> t.List[t.Dict[str, t.Any]]:
        async def sharded_request():
            i, j = 0, chunk_size
            tasks = []
            semaphore = asyncio.Semaphore(settings.MAX_NUM_REQUESTS_AT_A_TIME)
            number_of_chunks = self._get_number_of_chunks(adata, chunk_size=chunk_size)
            results = [[] for _ in range(number_of_chunks)]

            for chunk_index in range(number_of_chunks):
                chunk = adata[i:j, :]
                chunk_start_i = i
                chunk_end_i = i + len(chunk)
                self._print(
                    f"Submitting cell chunk #{chunk_index + 1:2.0f} ({chunk_start_i:5.0f}, {chunk_end_i:5.0f}) "
                    f"to CAS ..."
                )

                chunk_bytes = _io.adata_to_bytes(adata=chunk)
                async_sharded_request_callback = self.__get_async_sharded_request_callback(
                    results=results,
                    chunk_index=chunk_index,
                    chunk_start_i=chunk_start_i,
                    chunk_end_i=chunk_end_i,
                    semaphore=semaphore,
                    service_request_callback=request_callback,
                )
                async_sharded_request_task = asyncio.create_task(
                    async_sharded_request_callback(adata_bytes=chunk_bytes, **request_callback_kwargs)
                )
                tasks.append(async_sharded_request_task)
                i = j
                j += chunk_size

            await asyncio.wait(tasks)
            return functools.reduce(operator.iconcat, results, [])

        return asyncio.run(sharded_request())

    def __postprocess_sharded_response(
        self, query_response: t.List[t.Dict[str, t.Any]], adata: anndata.AnnData, query_item_list_key: str
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Postprocess results by matching the order of cells in the response with the order of cells in the input

        :param query_response: List of dictionaries with annotations for each of the cells from input adata
        :param adata: :class:`anndata.AnnData` instance to annotate
        :param query_item_list_key: Key in the dictionary that contains the list of items (e.g. annotation matches,
        search result items)

        :return: A list of dictionaries with annotations for each of the cells from input adata, ordered by the input
        """

        processed_response = []
        query_response_hash = {x["query_cell_id"]: x for x in query_response}

        num_unannotated_cells = 0

        for query_cell_id in adata.obs.index:
            try:
                query_item = query_response_hash[query_cell_id]
            except KeyError:
                query_item = {"query_cell_id": query_cell_id, query_item_list_key: []}
                num_unannotated_cells += 1

            processed_response.append(query_item)

        if num_unannotated_cells > 0:
            self._print(f"{num_unannotated_cells} cells were not processed by CAS")

        return processed_response

    def __postprocess_annotations(
        self, query_response: t.List[t.Dict[str, t.Any]], adata: anndata.AnnData
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Postprocess results by matching the order of cells in the response with the order of cells in the input
        """
        return self.__postprocess_sharded_response(
            query_response=query_response,
            adata=adata,
            query_item_list_key="matches",
        )

    def __postprocess_nearest_neighbor_search_response(
        self, query_response: t.List[t.Dict[str, t.Any]], adata: anndata.AnnData
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Postprocess nearest neighbor search response by matching the order of cells in the response with the order of
        cells in the input
        """
        return self.__postprocess_sharded_response(
            query_response=query_response,
            adata=adata,
            query_item_list_key="neighbors",
        )

    @staticmethod
    def __postprocess_query_cells_by_ids_response(
        query_response: t.List[t.Dict[str, t.Any]],
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Postprocess query cells by ids response by removing None values from cell metadata items in response

        :param query_response: List of dictionaries with annotations for each of the cells from input adata

        Returns:
        """
        processed_response = []

        for cell_metadata_item in query_response:
            cell_metadata_item_processed = {}

            for feature_name, value in cell_metadata_item.items():
                if value is None:
                    continue

                cell_metadata_item_processed[feature_name] = value

            processed_response.append(cell_metadata_item_processed)

        return processed_response

    def __validate_cells_under_quota(self, cell_count: int) -> None:
        """
        Validate the number of cells in the input data does not exceed remaining user quota

        :param cell_count: Number of cells in the input data
        """
        user_quota = self.cas_api_service.get_user_quota()
        if cell_count > user_quota["remaining_quota"]:
            raise exceptions.QuotaExceededError(
                f"Number of cells in the input data ({cell_count}) exceeds the user's remaining quota ({user_quota['remaining_quota']}).  "
                f"The user's quota will be reset to {user_quota['quota']} on {user_quota['quota_reset_date']}."
            )

    def __prepare_input_for_sharded_request(
        self,
        adata: anndata.AnnData,
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        cas_model_name: t.Optional[str] = None,
        feature_names_column_name: t.Optional[str] = None,
    ) -> anndata.AnnData:
        """
        Prepare input data for sharded request. Validates model name and input data, and sanitizes the input data

        :param adata: :class:`anndata.AnnData` instance to annotate
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``None``, which refers to the
            default selected model in the Cellarium backend. |br|
            `Default:` ``None``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        if cas_model_name not in self.allowed_models_list and cas_model_name is not None:
            raise ValueError(
                f"Model name '{cas_model_name}' is not in the list of allowed models. "
                f"Please use one of the following: {self.allowed_models_list} or `None` for the default model."
            )
        cas_model_name = self.default_model_name if cas_model_name is None else cas_model_name
        cas_model = self._model_name_obj_map[cas_model_name]
        cas_model_name = cas_model["model_name"]

        self._print(f"Cellarium CAS (Model ID: {cas_model_name})")
        self._print(f"Total number of input cells: {len(adata)}")

        self.__validate_cells_under_quota(cell_count=len(adata))

        return self.validate_and_sanitize_input_data(
            adata=adata,
            cas_model_name=cas_model_name,
            count_matrix_name=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )

    @deprecated(version="1.4.3", reason="Use :meth:`annotate_matrix_cell_type_summary_statistics_strategy` instead")
    @action_context_manager()
    def annotate_anndata(
        self,
        adata: "anndata.AnnData",
        chunk_size=CHUNK_SIZE_ANNOTATE_DEFAULT,
        cas_model_name: str = "default",
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        feature_names_column_name: t.Optional[str] = None,
        include_dev_metadata: bool = False,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Send an instance of :class:`anndata.AnnData` to the Cellarium Cloud backend for annotations. The function
        splits the ``adata`` into smaller chunks and asynchronously sends them to the backend API service. Each chunk is
        of equal size, except for the last one, which may be smaller. The backend processes these chunks in parallel.

        :param adata: :class:`anndata.AnnData` instance to annotate
        :param chunk_size: Size of chunks to split on
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``"default"``
            keyword, which refers to the default selected model in the Cellarium backend. |br|
            `Default:` ``"default"``
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``
        :param include_dev_metadata: Boolean indicating whether to include a breakdown of the number of cells
            by dataset

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        cas_model_name = self.default_model_name if cas_model_name == "default" else cas_model_name

        start = time.time()
        adata = self.__prepare_input_for_sharded_request(
            adata=adata,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )
        results = self.__async_sharded_request(
            adata=adata,
            chunk_size=chunk_size,
            request_callback=self.cas_api_service.async_annotate_cell_type_summary_statistics_strategy,
            request_callback_kwargs={
                "model_name": cas_model_name,
                "include_extended_output": include_dev_metadata,
            },
        )
        result = self.__postprocess_annotations(results, adata)
        self._print(f"Total wall clock time: {f'{time.time() - start:10.4f}'} seconds")
        self._render_feedback_link()
        return result

    @deprecated(version="1.4.3", reason="Use :meth:`annotate_matrix_cell_type_statistics_strategy` instead")
    @action_context_manager()
    def annotate_anndata_file(
        self,
        filepath: str,
        chunk_size=CHUNK_SIZE_ANNOTATE_DEFAULT,
        cas_model_name: str = "default",
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        feature_names_column_name: t.Optional[str] = None,
        include_dev_metadata: bool = False,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Read the 'h5ad' file into a :class:`anndata.AnnData` matrix and apply the :meth:`annotate_anndata` method to it.

        :param filepath: Filepath of the local :class:`anndata.AnnData` matrix
        :param chunk_size: Size of chunks to split on
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``"default"``
            keyword, which refers to the default selected model in the Cellarium backend. |br|
            `Default:` ``"default"``
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``
        :param include_dev_metadata: Boolean indicating whether to include a breakdown of the number of cells
            per dataset

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            adata = anndata.read_h5ad(filename=filepath)

        return self.annotate_anndata(
            adata=adata,
            chunk_size=chunk_size,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
            include_dev_metadata=include_dev_metadata,
        )

    @deprecated(version="1.4.3", reason="Use :meth:`annotate_matrix_with_cell_type_statistics_strategy` instead")
    @action_context_manager()
    def annotate_10x_h5_file(
        self,
        filepath: str,
        chunk_size: int = CHUNK_SIZE_ANNOTATE_DEFAULT,
        cas_model_name: str = "default",
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        feature_names_column_name: t.Optional[str] = None,
        include_dev_metadata: bool = False,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Parse the 10x 'h5' matrix and apply the :meth:`annotate_anndata` method to it.

        :param filepath: Filepath of the local 'h5' matrix
        :param chunk_size: Size of chunks to split on
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``"default"``
            keyword, which refers to the default selected model in the Cellarium backend. |br|
            `Default:` ``"default"``
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``
        :param include_dev_metadata: Boolean indicating whether to include a breakdown of the number of cells by dataset
        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        adata = _io.read_10x_h5(filepath)

        return self.annotate_anndata(
            adata=adata,
            chunk_size=chunk_size,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
            include_dev_metadata=include_dev_metadata,
        )

    @action_context_manager()
    def annotate_matrix_cell_type_summary_statistics_strategy(
        self,
        matrix: t.Union[str, anndata.AnnData],
        chunk_size=CHUNK_SIZE_ANNOTATE_DEFAULT,
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        include_extended_statistics: bool = True,
        cas_model_name: t.Optional[str] = None,
        feature_names_column_name: t.Optional[str] = None,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Send an instance of :class:`anndata.AnnData` to the Cellarium Cloud backend for annotations. The function
        splits the ``adata`` into smaller chunks and asynchronously sends them to the backend API service. Each chunk is
        of equal size, except for the last one, which may be smaller. The backend processes these chunks in parallel.

        :param matrix: Either path to a file (must be either `.h5` or `.h5ad`) or an :class:`anndata.AnnData` instance
            to annotate
        :param chunk_size: Size of chunks to split on
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param include_extended_statistics: Boolean indicating whether to include a breakdown of the number of cells
            by dataset
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``None``, which refers to the
            default selected model in the Cellarium backend. |br|
            `Default:` ``None``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        if isinstance(matrix, str):
            matrix = _io.read_h5_or_h5ad(filename=matrix)

        cas_model_name = self.default_model_name if cas_model_name is None else cas_model_name

        start = time.time()
        matrix = self.__prepare_input_for_sharded_request(
            adata=matrix,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )
        results = self.__async_sharded_request(
            adata=matrix,
            chunk_size=chunk_size,
            request_callback=self.cas_api_service.async_annotate_cell_type_summary_statistics_strategy,
            request_callback_kwargs={
                "model_name": cas_model_name,
                "include_extended_output": include_extended_statistics,
            },
        )
        result = self.__postprocess_annotations(results, matrix)
        self._print(f"Total wall clock time: {f'{time.time() - start:10.4f}'} seconds")
        self._render_feedback_link()
        return result

    @action_context_manager()
    def annotate_matrix_cell_type_ontology_aware_strategy(
        self,
        matrix: t.Union[str, anndata.AnnData],
        chunk_size=CHUNK_SIZE_ANNOTATE_DEFAULT,
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        cas_model_name: t.Optional[str] = None,
        feature_names_column_name: t.Optional[str] = None,
        prune_threshold: float = DEFAULT_PRUNE_THRESHOLD,
        weighting_prefactor: float = DEFAULT_WEIGHTING_PREFACTOR,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Send an instance of :class:`anndata.AnnData` to the Cellarium Cloud backend for annotations using ontology
        aware strategy . The function splits the ``adata`` into smaller chunks and asynchronously sends them to the
        backend API service. Each chunk is of equal size, except for the last one, which may be smaller. The backend
        processes these chunks in parallel.

        :param matrix: Either path to a file (must be either `.h5` or `.h5ad`) or an :class:`anndata.AnnData` instance
            to annotate
        :param chunk_size: Size of chunks to split on
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``None``, which refers to the
            default selected model in the Cellarium backend. |br|
            `Default:` ``None``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``
        :param prune_threshold: Threshold score for pruning the ontology graph in the output
        :param weighting_prefactor: Weighting prefactor for the weight calculation. A larger absolute value of the
            weighting_prefactor results in a steeper decay (weights drop off more quickly as distance increases),
            whereas a smaller absolute value results in a slower decay

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        if isinstance(matrix, str):
            matrix = _io.read_h5_or_h5ad(filename=matrix)

        cas_model_name = self.default_model_name if cas_model_name is None else cas_model_name

        start = time.time()
        matrix = self.__prepare_input_for_sharded_request(
            adata=matrix,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )
        results = self.__async_sharded_request(
            adata=matrix,
            chunk_size=chunk_size,
            request_callback=self.cas_api_service.async_annotate_cell_type_ontology_aware_strategy_anndata,
            request_callback_kwargs={
                "model_name": cas_model_name,
                "prune_threshold": prune_threshold,
                "weighting_prefactor": weighting_prefactor,
            },
        )
        result = self.__postprocess_annotations(results, matrix)
        self._print(f"Total wall clock time: {f'{time.time() - start:10.4f}'} seconds")
        self._render_feedback_link()
        return result

    @deprecated(version="1.4.3", reason="Use :meth:`search_matrix` instead")
    @action_context_manager()
    def search_anndata(
        self,
        adata: anndata.AnnData,
        chunk_size=CHUNK_SIZE_SEARCH_DEFAULT,
        cas_model_name: str = "default",
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        feature_names_column_name: t.Optional[str] = None,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Send an instance of :class:`anndata.AnnData` to the Cellarium Cloud backend for nearest neighbor search. The
        function splits the ``adata`` into smaller chunks and asynchronously sends them to the backend API service.
        Each chunk is of equal size, except for the last one, which may be smaller. The backend processes
        these chunks in parallel.

        :param adata: :class:`anndata.AnnData` instance to annotate
        :param chunk_size: Size of chunks to split on
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``"default"``
            keyword, which refers to the default selected model in the Cellarium backend. |br|
            `Default:` ``"default"``
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        if chunk_size > settings.MAX_CHUNK_SIZE_SEARCH_METHOD:
            raise ValueError("Chunk size greater than 500 not supported yet.")

        cas_model_name = self.default_model_name if cas_model_name == "default" else cas_model_name

        start = time.time()
        adata = self.__prepare_input_for_sharded_request(
            adata=adata,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )
        results = self.__async_sharded_request(
            adata=adata,
            chunk_size=chunk_size,
            request_callback=self.cas_api_service.async_nearest_neighbor_search,
            request_callback_kwargs={"model_name": cas_model_name},
        )
        result = self.__postprocess_nearest_neighbor_search_response(results, adata)
        self._print(f"Total wall clock time: {f'{time.time() - start:10.4f}'} seconds")
        return result

    @deprecated(version="1.4.3", reason="Use :meth:`search_matrix` instead")
    @action_context_manager()
    def search_10x_h5_file(
        self,
        filepath: str,
        chunk_size: int = CHUNK_SIZE_SEARCH_DEFAULT,
        cas_model_name: str = "default",
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        feature_names_column_name: t.Optional[str] = None,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Parse the 10x 'h5' matrix and apply the :meth:`search_anndata` method to it.

        :param filepath: Filepath of the local 'h5' matrix
        :param chunk_size: Size of chunks to split on
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``"default"``
            keyword, which refers to the default selected model in the Cellarium backend. |br|
            `Default:` ``"default"``
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        adata = _io.read_10x_h5(filepath)

        return self.search_anndata(
            adata=adata,
            chunk_size=chunk_size,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )

    @action_context_manager()
    def search_matrix(
        self,
        matrix: t.Union[str, anndata.AnnData],
        chunk_size: int = CHUNK_SIZE_SEARCH_DEFAULT,
        count_matrix_input: constants.CountMatrixInput = constants.CountMatrixInput.X,
        feature_ids_column_name: str = "index",
        cas_model_name: t.Optional[str] = None,
        feature_names_column_name: t.Optional[str] = None,
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Send an instance of :class:`anndata.AnnData` to the Cellarium Cloud backend for nearest neighbor search. The
        function splits the ``adata`` into smaller chunks and asynchronously sends them to the backend API service.
        Each chunk is of equal size, except for the last one, which may be smaller. The backend processes
        these chunks in parallel.

        :param matrix: Either path to a file (must be either `.h5` or `.h5ad`) or an :class:`anndata.AnnData` instance
            to annotate
        :param chunk_size: Size of chunks to split on
        :param count_matrix_input:  Where to obtain a feature expression count matrix from. |br|
            `Allowed Values: Choices from enum :class:`cellarium.cas.constants.CountMatrixInput` |br|
            `Default:` ``"CountMatrixInput.X"``
        :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``"index"``
        :param cas_model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``None``, which refers to the
            default selected model in the Cellarium backend. |br|
            `Default:` ``None``
        :param feature_names_column_name: Column name where to obtain feature names (symbols).
            feature names wouldn't be mapped if value is ``None`` |br|
            `Allowed Values:` A value from ``adata.var.columns`` or ``"index"`` keyword, which refers to index
            column. |br|
            `Default:` ``None``

        :return: A list of dictionaries with annotations for each of the cells from input adata
        """
        if isinstance(matrix, str):
            matrix = _io.read_h5_or_h5ad(filename=matrix)

        if chunk_size > settings.MAX_CHUNK_SIZE_SEARCH_METHOD:
            raise ValueError(f"Chunk size greater than `{settings.MAX_CHUNK_SIZE_SEARCH_METHOD}` not supported yet.")

        cas_model_name = self.default_model_name if cas_model_name is None else cas_model_name

        start = time.time()
        matrix = self.__prepare_input_for_sharded_request(
            adata=matrix,
            cas_model_name=cas_model_name,
            count_matrix_input=count_matrix_input,
            feature_ids_column_name=feature_ids_column_name,
            feature_names_column_name=feature_names_column_name,
        )
        results = self.__async_sharded_request(
            adata=matrix,
            chunk_size=chunk_size,
            request_callback=self.cas_api_service.async_nearest_neighbor_search,
            request_callback_kwargs={"model_name": cas_model_name},
        )
        result = self.__postprocess_nearest_neighbor_search_response(results, matrix)
        self._print(f"Total wall clock time: {f'{time.time() - start:10.4f}'} seconds")
        self._render_feedback_link()
        return result

    @action_context_manager()
    def query_cells_by_ids(
        self, cell_ids: t.List[int], model_name: t.Optional[str], metadata_feature_names: t.List[str] = None
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Query cells by their ids from a single anndata file with Cellarium CAS. Input file should be validated and
        sanitized according to the model schema.

        :param cell_ids: List of cell ids to query
        :param model_name: Model name to use for annotation. |br|
            `Allowed Values:` Model name from the :attr:`allowed_models_list` list or ``None``
            keyword, which refers to the default selected model in the Cellarium backend. |br|
            `Default:` ``None``
        :param metadata_feature_names: List of metadata feature names to include in the response. |br|

        :return: List of cells with metadata
        """
        model_name = self.default_model_name if model_name is None else model_name
        results = self.cas_api_service.query_cells_by_ids(
            cell_ids=cell_ids,
            model_name=model_name,
            metadata_feature_names=metadata_feature_names,
        )
        cells = self.__postprocess_query_cells_by_ids_response(query_response=results)
        self._render_feedback_link()
        return cells

    def validate_model_name(self, model_name: t.Optional[str] = None) -> None:
        """
        Validate if the model name provided is valid

        :param model_name: Model name to check
        :raises: ValueError if model name is not valid
        """
        if model_name not in self.allowed_models_list and model_name is not None:
            raise ValueError(
                f"Model name '{model_name}' is not in the list of allowed models. "
                f"Please use one of the following: {self.allowed_models_list} or `None` for the default model."
            )



import os
import sys
import tempfile
import typing as t
import warnings
from contextlib import contextmanager

import anndata
import h5py
import numpy as np

if t.TYPE_CHECKING:
    import pathlib


def _collect_datasets(dsets: t.Dict, group: "h5py.Group"):
    for k, v in group.items():
        if isinstance(v, h5py.Dataset):
            dsets[k] = v[()]
        else:
            _collect_datasets(dsets, v)


def _read_legacy_10x_h5(filename, *, genome=None):
    """
    Read hdf5 file from Cell Ranger v2 or earlier versions.
    """
    with h5py.File(str(filename), "r") as f:
        try:
            children = list(f.keys())
            if not genome:
                if len(children) > 1:
                    raise ValueError(
                        f"'{filename}' contains more than one genome. For legacy 10x h5 "
                        "files you must specify the genome if more than one is present. "
                        f"Available genomes are: {children}"
                    )
                genome = children[0]
            elif genome not in children:
                raise ValueError(
                    f"Could not find genome '{genome}' in '{filename}'. " f"Available genomes are: {children}"
                )

            dsets = {}
            _collect_datasets(dsets, f[genome])

            # AnnData works with csr matrices
            # 10x stores the transposed data, so we do the transposition right away
            from scipy.sparse import csr_matrix

            M, N = dsets["shape"]
            data = dsets["data"]
            if dsets["data"].dtype == np.dtype("int32"):
                data = dsets["data"].view("float32")
                data[:] = dsets["data"]
            matrix = csr_matrix(
                (data, dsets["indices"], dsets["indptr"]),
                shape=(N, M),
            )
            # the csc matrix is automatically the transposed csr matrix
            # as scanpy expects it, so, no need for a further transposition
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                adata = anndata.AnnData(
                    matrix,
                    obs=dict(obs_names=dsets["barcodes"].astype(str)),
                    var=dict(
                        var_names=dsets["gene_names"].astype(str),
                        gene_ids=dsets["genes"].astype(str),
                    ),
                )

            return adata
        except KeyError:
            raise Exception("File is missing one or more required datasets.")


def _read_v3_10x_h5(filename: str):
    """
    Read hdf5 file from Cell Ranger v3 or later versions.
    """
    with h5py.File(str(filename), "r") as f:
        try:
            dsets = {}
            _collect_datasets(dsets, f["matrix"])

            from scipy.sparse import csr_matrix

            M, N = dsets["shape"]
            data = dsets["data"]
            if dsets["data"].dtype == np.dtype("int32"):
                data = dsets["data"].view("float32")
                data[:] = dsets["data"]
            matrix = csr_matrix(
                (data, dsets["indices"], dsets["indptr"]),
                shape=(N, M),
            )
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                adata = anndata.AnnData(
                    matrix,
                    obs=dict(obs_names=dsets["barcodes"].astype(str)),
                    var=dict(
                        var_names=dsets["id"].astype(str),
                        gene_name=dsets["name"].astype(str),
                        feature_type=dsets["feature_type"].astype(str),
                        genome=dsets["genome"].astype(str),
                    ),
                )
            return adata
        except KeyError:
            raise Exception("File is missing one or more required datasets.")


def read_10x_h5(
    filename: t.Union[str, "pathlib.Path"],
    genome: t.Optional[str] = None,
) -> "anndata.AnnData":
    """
     Read 10x-Genomics-formatted hdf5 file.
     This method is borrowed from scanpy, please refer to scanpy library:
     https://github.com/scverse/scanpy/blob/1.6.x/scanpy/readwrite.py#LL138C10-L138C10.

    :param filename:  Path to a 10x hdf5 file.
    :param genome: Filter expression to genes within this genome. For legacy 10x h5 files, this must be provided
        if the data contains more than one genome.
    :return: Annotated data matrix, where observations/cells are named by their barcode and variables/genes by
        gene name. Stores the following information:
            :attr:`~anndata.AnnData.X`
                The data matrix is stored
            :attr:`~anndata.AnnData.obs_names`
                Cell names
            :attr:`~anndata.AnnData.var_names`
                Gene names
            :attr:`~anndata.AnnData.var`\\ `['gene_ids']`
                Gene IDs
            :attr:`~anndata.AnnData.var`\\ `['feature_types']`
                Feature types
    """
    with h5py.File(str(filename), "r") as f:
        v3 = "/matrix" in f
    if v3:
        adata = _read_v3_10x_h5(filename)
        if adata.is_view:
            adata = adata.copy()
    else:
        adata = _read_legacy_10x_h5(filename, genome=genome)
    return adata


def read_h5_or_h5ad(filename: str) -> anndata.AnnData:
    """
    Read an `h5` or `h5ad` file and return an AnnData object.

    :param filename: Path to the h5 or h5ad file.
    :return: An AnnData object.

    Usage example:

    .. code-block:: python

        import anndata

        adata = read_h5_or_h5ad("path/to/file.h5ad")
    """
    if filename.endswith(".h5"):
        return read_10x_h5(filename)
    elif filename.endswith(".h5ad"):
        return anndata.read_h5ad(filename)
    else:
        raise ValueError("File should be either a `h5` or `h5ad` file.")


def adata_to_bytes(adata: "anndata.AnnData", compression: str = "gzip") -> bytes:
    """
    Convert an :class:`anndata.AnnData` object to a byte object.

    :param adata: The AnnData object to be serialized.
    :param compression: The compression type to be used. Should be any compression type supported by h5py.
        `Default:` ``"gzip"``
    :return: Serialized bytes objects of the AnnData object.

    Usage example:

    .. code-block:: python

        import anndata

        adata = anndata.AnnData(X)
        adata_bytes = adata_to_bytes(adata)
    """

    with tempfile.NamedTemporaryFile(suffix=".h5ad") as temp_file:
        adata.write(temp_file.name, compression=compression)
        temp_file.seek(0)
        return temp_file.read()


@contextmanager
def suppress_stderr():
    original_stderr = sys.stderr
    sys.stderr = open(os.devnull, "w")
    try:
        yield
    finally:
        sys.stderr.close()
        sys.stderr = original_stderr



import typing as t

import anndata
import numpy as np
import pandas as pd
import scipy.sparse as sp

from cellarium.cas import constants, exceptions
from cellarium.cas.preprocessing import callbacks


def _get_adata_var_index_or_by_column(adata: anndata.AnnData, var_column_name: str) -> t.List[str]:
    if var_column_name == "index":
        return adata.var.index.tolist()

    return adata.var[var_column_name].values.tolist()


def validate(
    adata: anndata.AnnData,
    cas_feature_schema_list: t.List,
    feature_ids_column_name: str,
) -> None:
    """
    Validate input `anndata.AnnData` instance in concordance with feature schema `cas_feature_schema_list`.
    Raise `exceptions.DataValidationError` with number of missing features and number of extra features
    that were present in dataset.

    :param adata: Instance to validate
    :param cas_feature_schema_list: List of features to be validated with
    :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. Default `index`.
    """
    if feature_ids_column_name != "index" and feature_ids_column_name not in adata.var.columns.values:
        raise ValueError(
            "`feature_ids_column_name` should have a value of either 'index' "
            "or be present as a column in the `adata.var` object."
        )

    adata_feature_schema_list = _get_adata_var_index_or_by_column(adata=adata, var_column_name=feature_ids_column_name)

    cas_feature_schema_set = set(cas_feature_schema_list)
    adata_feature_schema_set = set(adata_feature_schema_list)

    if adata_feature_schema_list == cas_feature_schema_list:
        return

    missing_features = len(cas_feature_schema_set - adata_feature_schema_set)
    extra_features = len(adata_feature_schema_set - cas_feature_schema_set)

    raise exceptions.DataValidationError(missing_features=missing_features, extra_features=extra_features)


def sanitize(
    adata: anndata.AnnData,
    cas_feature_schema_list: t.List[str],
    count_matrix_input: constants.CountMatrixInput,
    feature_ids_column_name: str,
    feature_names_column_name: t.Optional[str] = None,
) -> anndata.AnnData:
    """
    Cellarium CAS sanitizing script. Returns a new `anndata.AnnData` instance, based on the feature expression
    matrix of the input instance. Extra features get omitted. Missing features get filled with zeros.

    :param adata: Instance to sanitize
    :param cas_feature_schema_list: List of Ensembl feature ids to rely on.
    :param count_matrix_input: Where to obtain a feature expression count matrix from. Choice of: 'X', 'raw.X'
    :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. Default `index`.
    :param feature_names_column_name: Column name where to obtain feature names. If not provided, no feature names
        should be mapped |br|
    `Default`: ``None``
    :return: `anndata.AnnData` instance that is ready to use with CAS
    """
    if feature_ids_column_name != "index" and feature_ids_column_name not in adata.var.columns.values:
        raise ValueError(
            "`feature_ids_column_name` should have a value of either 'index' "
            "or be present as a column in the `adata.var` object."
        )
    if feature_names_column_name not in {"index", None} and feature_names_column_name not in adata.var.columns.values:
        raise ValueError(
            "`feature_ids_name_column_name` should have a value of either 'index' "
            "or be present as a column in the `adata.var` object."
        )

    callbacks.pre_sanitize_callback(adata=adata, count_matrix_input=count_matrix_input)

    adata_feature_schema_list = _get_adata_var_index_or_by_column(adata=adata, var_column_name=feature_ids_column_name)
    original_obs_ids = adata.obs.index.values

    cas_feature_schema_set = set(cas_feature_schema_list)
    adata_feature_schema_set = set(adata_feature_schema_list)
    feature_id_intersection = adata_feature_schema_set.intersection(cas_feature_schema_set)

    cas_feature_id_map = {feature_id: index for index, feature_id in enumerate(cas_feature_schema_list)}
    adata_feature_id_map = {feature_id: index for index, feature_id in enumerate(adata_feature_schema_list)}
    feature_id_intersection_cas_indices = list(map(cas_feature_id_map.get, feature_id_intersection))
    feature_id_intersection_adata_indices = list(map(adata_feature_id_map.get, feature_id_intersection))

    n_cells = adata.shape[0]
    n_features = len(cas_feature_schema_list)
    input_matrix = adata.X if count_matrix_input == constants.CountMatrixInput.X else adata.raw.X

    # Translate the columns from one matrix to another, convert to COO format to make this efficient.
    col_trans = np.zeros(n_features, dtype=int)
    for i, k in enumerate(feature_id_intersection_cas_indices):
        col_trans[i] = k
    vals = input_matrix.tocsc()[:, feature_id_intersection_adata_indices]
    vals = vals.tocoo()
    new_col = col_trans[vals.col]
    result_matrix = sp.coo_matrix((vals.data, (vals.row, new_col)), shape=(n_cells, n_features))
    del col_trans, vals, new_col

    # Create `obs` index
    obs = adata.obs.copy()
    obs.index = original_obs_ids

    var_df_data = {}
    if feature_names_column_name is not None:
        adata_feature_name_list = _get_adata_var_index_or_by_column(
            adata=adata, var_column_name=feature_names_column_name
        )
        gene_id_to_gene_symbol_map = {
            gene_id: gene_symbol for gene_symbol, gene_id in zip(adata_feature_name_list, adata_feature_schema_list)
        }

        cas_feature_name_list = [
            gene_id_to_gene_symbol_map[gene_id] if gene_id in gene_id_to_gene_symbol_map else "N/A"
            for gene_id in cas_feature_schema_list
        ]

        var_df_data["feature_name"] = cas_feature_name_list

    var_df = pd.DataFrame(index=cas_feature_schema_list, data=var_df_data)
    return anndata.AnnData(
        result_matrix.tocsr(),
        obs=obs,
        obsm=adata.obsm,
        var=var_df,
    )



from ..version import get_version_environment
from .base import *  # noqa

VERSION_ENVIRONMENT = get_version_environment()

if VERSION_ENVIRONMENT == "development" or VERSION_ENVIRONMENT == "test":
    from .development import *  # noqa
else:
    from .production import *  # noqa



import logging
import os
import tempfile
import typing as t
from collections import OrderedDict
from logging import log

import dash_bootstrap_components as dbc
import numpy as np
import plotly.graph_objects as go
from anndata import AnnData
from Bio import Phylo
from dash import Dash, State, dcc, html
from dash.dependencies import Input, Output
from dash.development.base_component import Component
from plotly.express.colors import sample_colorscale

from cellarium.cas.postprocessing import (
    CAS_CL_SCORES_ANNDATA_OBSM_KEY,
    CellOntologyScoresAggregationDomain,
    CellOntologyScoresAggregationOp,
    convert_aggregated_cell_ontology_scores_to_rooted_tree,
    generate_phyloxml_from_scored_cell_ontology_tree,
    get_aggregated_cas_ontology_aware_scores,
    get_obs_indices_for_cluster,
    insert_cas_ontology_aware_response_into_adata,
)
from cellarium.cas.postprocessing.cell_ontology import CL_EUKARYOTIC_CELL_ROOT_NODE, CellOntologyCache
from cellarium.cas.visualization._components.circular_tree_plot import CircularTreePlot
from cellarium.cas.visualization.ui_utils import ConfigValue, find_and_kill_process

# cell type ontology terms (and all descendents) to hide from the visualization
DEFAULT_HIDDEN_CL_NAMES_SET = {}

# header component ID -> default title mapping for the panels
DEFAULT_PANEL_TITLES = {
    "cell-selection-title-tree": "Cell Type Ontology View",
    "cell-selection-title-umap": "UMAP View",
}


class DomainSelectionConstants:
    NONE = 0
    USER_SELECTION = 1
    SEPARATOR = 2


# cell type ontology terms to always show as text labels in the visualization
DEFAULT_SHOWN_CL_NAMES_SET = {
    "CL_0000236",
    "CL_0000084",
    "CL_0000789",
    "CL_0000798",
    "CL_0002420",
    "CL_0002419",
    "CL_0000786",
    "CL_0000576",
    "CL_0001065",
    "CL_0000451",
    "CL_0000094",
    "CL_0000235",
    "CL_0000097",
    "CL_0000814",
    "CL_0000827",
    "CL_0000066",
    "CL_0000163",
    "CL_0000151",
    "CL_0000064",
    "CL_0000322",
    "CL_0000076",
    "CL_0005006",
    "CL_0000148",
    "CL_0000646",
    "CL_0009004",
    "CL_0000115",
    "CL_0000125",
    "CL_0002319",
    "CL_0000187",
    "CL_0000057",
    "CL_0008034",
    "CL_0000092",
    "CL_0000058",
    "CL_0000060",
    "CL_0000136",
    "CL_0000499",
    "CL_0000222",
    "CL_0007005",
    "CL_0000039",
    "CL_0000019",
    "CL_0000223",
    "CL_0008019",
    "CL_0005026",
    "CL_0000182",
    "CL_0000023",
    "CL_0000679",
    "CL_0000126",
    "CL_0000540",
    "CL_0000127",
    "CL_0011005",
}


class CASCircularTreePlotUMAPDashApp:
    """
    A Dash app for visualizing the results of a Cellarium CAS cell type ontology-aware analysis.

    :param adata: The AnnData object containing the cell type ontology-aware analysis results.
    :param cas_ontology_aware_response: The response from the Cellarium CAS cell type ontology-aware analysis. |br|
        `Default:` ``None``
    :param cluster_label_obs_column: The name of the observation column containing the cluster labels. |br|
        `Default:` ``None``
    :param aggregation_op: The aggregation operation to apply to the cell type ontology-aware scores. |br|
        `Default:` ``CellOntologyScoresAggregationOp.MEAN``
    :param aggregation_domain: The domain over which to aggregate the cell type ontology-aware scores. |br|
        `Default:` ``CellOntologyScoresAggregationDomain.OVER_THRESHOLD``
    :param score_threshold: The threshold for the cell type ontology-aware scores. |br|
        `Default:` ``0.05``
    :param min_cell_fraction: The minimum fraction of cells that must have a cell type ontology-aware score above the threshold. |br|
        `Default:` ``0.01``
    :param umap_marker_size: The size of the markers in the UMAP scatter plot. |br|
        `Default:` ``3.0``
    :param umap_padding: The padding to apply to the UMAP scatter plot bounds. |br|
        `Default:` ``0.15``
    :param umap_min_opacity: The minimum opacity for the UMAP scatter plot markers. |br|
        `Default:` ``0.1``
    :param umap_max_opacity: The maximum opacity for the UMAP scatter plot markers. |br|
        `Default:` ``1.0``
    :param umap_inactive_cell_color: The color for inactive cells in the UMAP scatter plot. |br|
        `Default:` ``"rgb(180,180,180)"``
    :param umap_inactive_cell_opacity: The opacity for inactive cells in the UMAP scatter plot. |br|
        `Default:` ``0.5``
    :param umap_active_cell_color: The color for active cells in the UMAP scatter plot. |br|
        `Default:` ``"rgb(250,50,50)"``
    :param umap_default_cell_color: The default color for cells in the UMAP scatter plot. |br|
        `Default:` ``"rgb(180,180,180)"``
    :param umap_default_opacity: The default opacity for cells in the UMAP scatter plot. |br|
        `Default:` ``0.9``
    :param circular_tree_plot_linecolor: The line color for the circular tree plot. |br|
        `Default:` ``"rgb(200,200,200)"``
    :param circular_tree_start_angle: The start angle for the circular tree plot. |br|
        `Default:` ``180``
    :param circular_tree_end_angle: The end angle for the circular tree plot. |br|
        `Default:` ``360``
    :param figure_height: The height of the figures in the Dash app. |br|
        `Default:` ``400``
    :param hidden_cl_names_set: The set of cell type ontology terms to hide from the visualization. |br|
        `Default:` ``DEFAULT_HIDDEN_CL_NAMES_SET``
    :param shown_cl_names_set: The set of cell type ontology terms to always show as text labels in the
        visualization. |br|
        `Default:` ``DEFAULT_SHOWN_CL_NAMES_SET``
    :param score_colorscale: The colorscale to use for the cell type ontology-aware scores. |br|
        `Default:` ``"Viridis"``

    Example:
    ________
        >>> from cellarium.cas._io import suppress_stderr
        >>> from cellarium.cas.visualization import CASCircularTreePlotUMAPDashApp
        >>> DASH_SERVER_PORT = 8050
        >>> adata = ... # get your matrix
        >>> cas_ontology_aware_response = cas.annotate_matrix_cell_type_ontology_aware_strategy(
        >>>     matrix=adata,
        >>>     chunk_size=500
        >>> )
        >>> with suppress_stderr():
        >>>     CASCircularTreePlotUMAPDashApp(
        >>>         adata,
        >>>         cas_ontology_aware_response,
        >>>         cluster_label_obs_column="cluster_label",
        >>>     ).run(port=DASH_SERVER_PORT, debug=False, jupyter_width="100%")
    """

    ALL_CELLS_DOMAIN_KEY = "all cells"
    CLUSTER_PREFIX_DOMAIN_KEY = "cluster "

    def __init__(
        self,
        adata: AnnData,
        cas_ontology_aware_response: list,
        cluster_label_obs_column: t.Optional[str] = None,
        aggregation_op: CellOntologyScoresAggregationOp = CellOntologyScoresAggregationOp.MEAN,
        aggregation_domain: CellOntologyScoresAggregationDomain = CellOntologyScoresAggregationDomain.OVER_THRESHOLD,
        score_threshold: float = 0.05,
        min_cell_fraction: float = 0.01,
        umap_marker_size: float = 3.0,
        umap_padding: float = 0.15,
        umap_min_opacity: float = 0.1,
        umap_max_opacity: float = 1.0,
        umap_inactive_cell_color: str = "rgb(180,180,180)",
        umap_inactive_cell_opacity: float = 0.5,
        umap_active_cell_color: str = "rgb(250,50,50)",
        umap_default_cell_color: str = "rgb(180,180,180)",
        umap_default_opacity: float = 0.9,
        circular_tree_plot_linecolor: str = "rgb(200,200,200)",
        circular_tree_start_angle: int = 180,
        circular_tree_end_angle: int = 360,
        figure_height: int = 400,
        hidden_cl_names_set: set[str] = DEFAULT_HIDDEN_CL_NAMES_SET,
        shown_cl_names_set: set[str] = DEFAULT_SHOWN_CL_NAMES_SET,
        score_colorscale: t.Union[str, list] = "Viridis",
    ):
        self.adata = adata
        self.aggregation_op = aggregation_op
        self.aggregation_domain = aggregation_domain
        self.score_threshold = ConfigValue(score_threshold)
        self.min_cell_fraction = ConfigValue(min_cell_fraction)
        self.umap_min_opacity = umap_min_opacity
        self.umap_max_opacity = umap_max_opacity
        self.umap_marker_size = umap_marker_size
        self.umap_padding = umap_padding
        self.umap_inactive_cell_color = umap_inactive_cell_color
        self.umap_inactive_cell_opacity = umap_inactive_cell_opacity
        self.umap_active_cell_color = umap_active_cell_color
        self.umap_default_cell_color = umap_default_cell_color
        self.umap_default_opacity = umap_default_opacity
        self.circular_tree_plot_linecolor = circular_tree_plot_linecolor
        self.circular_tree_start_angle = circular_tree_start_angle
        self.circular_tree_end_angle = circular_tree_end_angle
        self.height = figure_height
        self.hidden_cl_names_set = hidden_cl_names_set
        self.shown_cl_names_set = shown_cl_names_set
        self.score_colorscale = score_colorscale

        assert "X_umap" in adata.obsm, "UMAP coordinates not found in adata.obsm['X_umap']"

        # setup cell domains
        self.cell_domain_map = OrderedDict()
        self.cell_domain_map[self.ALL_CELLS_DOMAIN_KEY] = np.arange(adata.n_obs)
        if cluster_label_obs_column is not None:
            assert cluster_label_obs_column in adata.obs
            for cluster_label in adata.obs[cluster_label_obs_column].cat.categories:
                self.cell_domain_map[self.CLUSTER_PREFIX_DOMAIN_KEY + cluster_label] = get_obs_indices_for_cluster(
                    adata, cluster_label_obs_column, cluster_label
                )

        # default cell domain
        self.selected_cell_domain_key = ConfigValue(DomainSelectionConstants.NONE)
        # Selected cells (from UMAP chart)
        self.selected_cells = []
        # Selected cell class (from tree diagram)
        self.selected_cl_name = None

        # instantiate the cell type ontology cache
        self.cl = CellOntologyCache()

        # insert CA ontology-aware response into adata
        insert_cas_ontology_aware_response_into_adata(cas_ontology_aware_response, adata, self.cl)

        # instantiate the Dash app
        self.app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.BOOTSTRAP])
        self.server = self.app.server
        self.app.layout = self._create_layout()
        self._setup_initialization()
        self._setup_callbacks()

    def run(self, port: int = 8050, **kwargs):
        log(logging.INFO, "Starting Dash application on port {port}...")
        try:
            self.app.run_server(port=port, jupyter_mode="inline", jupyter_height=self.height + 100, **kwargs)
        except OSError:  # Dash raises OSError if the port is already in use
            find_and_kill_process(port)
            self.app.run_server(port=port, jupyter_mode="inline", jupyter_height=self.height + 100, **kwargs)

    def _instantiate_circular_tree_plot(self) -> CircularTreePlot:
        # reduce scores over the provided cells
        selected_cells = self._get_effective_selected_cells()
        aggregated_scores = get_aggregated_cas_ontology_aware_scores(
            self.adata,
            obs_indices=(
                self.cell_domain_map[self.ALL_CELLS_DOMAIN_KEY] if len(selected_cells) == 0 else selected_cells
            ),
            aggregation_op=self.aggregation_op,
            aggregation_domain=self.aggregation_domain,
            threshold=self.score_threshold.get(),
        )

        # generate a Phylo tree
        rooted_tree = convert_aggregated_cell_ontology_scores_to_rooted_tree(
            aggregated_scores=aggregated_scores,
            cl=self.cl,
            root_cl_name=CL_EUKARYOTIC_CELL_ROOT_NODE,
            min_fraction=self.min_cell_fraction.get(),
            hidden_cl_names_set=self.hidden_cl_names_set,
        )
        phyloxml_string = generate_phyloxml_from_scored_cell_ontology_tree(
            rooted_tree, "Scored cell type ontology tree", self.cl, indent=3
        )

        with tempfile.NamedTemporaryFile(delete=False, mode="w+t") as temp_file:
            temp_file_name = temp_file.name
            temp_file.write(phyloxml_string)
            temp_file.flush()

        try:
            phyloxml_tree = Phylo.read(temp_file_name, "phyloxml")
        finally:
            os.remove(temp_file_name)

        return CircularTreePlot(
            tree=phyloxml_tree,
            score_colorscale=self.score_colorscale,
            linecolor=self.circular_tree_plot_linecolor,
            start_angle=self.circular_tree_start_angle,
            end_angle=self.circular_tree_end_angle,
            shown_cl_names_set=self.shown_cl_names_set,
        )

    def _get_padded_umap_bounds(self, umap_padding: float) -> t.Tuple[float, float, float, float]:
        actual_min_x = np.min(self.adata.obsm["X_umap"][:, 0])
        actual_max_x = np.max(self.adata.obsm["X_umap"][:, 0])
        actual_min_y = np.min(self.adata.obsm["X_umap"][:, 1])
        actual_max_y = np.max(self.adata.obsm["X_umap"][:, 1])
        padded_min_x = actual_min_x - umap_padding * (actual_max_x - actual_min_x)
        padded_max_x = actual_max_x + umap_padding * (actual_max_x - actual_min_x)
        padded_min_y = actual_min_y - umap_padding * (actual_max_y - actual_min_y)
        padded_max_y = actual_max_y + umap_padding * (actual_max_y - actual_min_y)

        return padded_min_x, padded_max_x, padded_min_y, padded_max_y

    def _get_scores_for_cl_name(self, cl_name: str) -> np.ndarray:
        cl_index = self.cl.cl_names_to_idx_map[cl_name]
        return self.adata.obsm[CAS_CL_SCORES_ANNDATA_OBSM_KEY][:, cl_index].toarray().flatten()

    def _get_scatter_plot_opacity_from_scores(self, scores: np.ndarray) -> np.ndarray:
        min_score = np.min(scores)
        max_score = np.max(scores)
        normalized_scores = (scores - min_score) / (1e-6 + max_score - min_score)
        return np.maximum(
            scores, self.umap_min_opacity + (self.umap_max_opacity - self.umap_min_opacity) * normalized_scores
        )

    def _create_layout(self):
        layout = html.Div(
            [
                dbc.Row(dbc.Col(className="gr-spacer", width=12)),
                dbc.Row(
                    dbc.Col(
                        [
                            html.H3(self._render_breadcrumb(), id="selected-domain-label", className="gr-breadcrumb"),
                            html.Div(
                                [
                                    dbc.ButtonGroup(
                                        [
                                            dbc.Button(
                                                html.I(className="bi bi-gear-fill"),
                                                id="settings-button",
                                                n_clicks=0,
                                                size="sm",
                                            ),
                                        ]
                                    )
                                ],
                                className="gr-settings-buttons",
                            ),
                        ],
                        className="gr-title",
                        width=12,
                    )
                ),
                dbc.Row(
                    self._render_cell_selection_panes(),
                    id="panel-titles",
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            html.Div(
                                [
                                    dcc.Graph(
                                        id="circular-tree-plot",
                                        style={
                                            "width": "100%",
                                            "display": "inline-block",
                                            "height": f"{self.height}px",
                                        },
                                        config={"scrollZoom": True},
                                    ),
                                ]
                            ),
                            width=6,
                        ),
                        dbc.Col(
                            html.Div(
                                [
                                    dcc.Graph(
                                        id="umap-scatter-plot",
                                        style={
                                            "width": "100%",
                                            "display": "inline-block",
                                            "height": f"{self.height-10}px",
                                        },
                                        # Zoom is very choppy with this enabled.  Users should use selection zoom
                                        config={"scrollZoom": False},
                                    ),
                                ]
                            ),
                            width=6,
                        ),
                    ]
                ),
                dbc.Offcanvas(
                    id="settings-pane", title="Settings", is_open=False, children=self._render_closed_settings_pane()
                ),
                html.Div(id="init", style={"display": "none"}),
                html.Div(id="no-action", style={"display": "none"}),
            ],
        )

        return layout

    def _initialize_umap_scatter_plot(self) -> go.Figure:
        # calculate static bounds for the UMAP scatter plot
        self.umap_min_x, self.umap_max_x, self.umap_min_y, self.umap_max_y = self._get_padded_umap_bounds(
            self.umap_padding
        )

        fig = go.Figure()

        color = self.umap_default_cell_color

        selected_cells = self._get_effective_selected_cells()

        if len(selected_cells) > 0:
            color = [self.umap_inactive_cell_color] * self.adata.n_obs
            for i_obs in selected_cells:
                color[i_obs] = self.umap_active_cell_color

        fig.add_trace(
            go.Scatter(
                x=self.adata.obsm["X_umap"][:, 0],
                y=self.adata.obsm["X_umap"][:, 1],
                mode="markers",
                marker=dict(
                    color=color,
                    size=self.umap_marker_size,
                    opacity=self.umap_default_opacity,
                ),
                unselected=dict(marker=dict(opacity=self.umap_default_opacity)),
            )
        )
        self._update_umap_scatter_plot_layout(fig)
        self._umap_scatter_plot_figure = fig
        return fig

    def _initialize_circular_tree_plot(self) -> go.Figure:
        self.circular_tree_plot = self._instantiate_circular_tree_plot()
        fig = self.circular_tree_plot.plotly_figure
        return fig

    def _setup_initialization(self):
        @self.app.callback(Output("umap-scatter-plot", "figure"), Input("init", "children"))
        def _initialize_umap_scatter_plot(init):
            return self._initialize_umap_scatter_plot()

        @self.app.callback(Output("circular-tree-plot", "figure"), Input("init", "children"))
        def _initialize_circular_tree_plot(init):
            return self._initialize_circular_tree_plot()

    def _update_umap_scatter_plot_layout(self, umap_scatter_plot_fig):
        umap_scatter_plot_fig.update_layout(
            # uirevision is needed to maintain pan/zoom state.  It must be updated to trigger a refresh
            uirevision="true",
            plot_bgcolor="white",
            margin=dict(l=0, r=25, t=50, b=0),
            xaxis=dict(
                title="UMAP 1",
                showgrid=False,
                zeroline=False,  # Keep zero line enabled
                zerolinecolor="black",
                range=[self.umap_min_x, self.umap_max_x],  # Set x-axis limits
                showline=True,  # Show axis line
                linecolor="black",
                linewidth=1,
                tickmode="linear",
                tick0=-10,
                dtick=5,
            ),
            yaxis=dict(
                title="UMAP 2",
                showgrid=False,
                zeroline=False,  # Keep zero line enabled
                zerolinecolor="black",
                range=[self.umap_min_y, self.umap_max_y],  # Set y-axis limits
                showline=True,  # Show axis line
                linecolor="black",
                linewidth=1,
                tickmode="linear",
                tick0=-10,
                dtick=5,
            ),
            dragmode="pan",
        )

    def _render_breadcrumb(self) -> Component:
        selected_cells = self._get_effective_selected_cells()
        if len(selected_cells) == 0 and self.selected_cell_domain_key.get() == DomainSelectionConstants.NONE:
            label = "Viewing results for all cells"
            show_clear = False
        elif (
            len(selected_cells) == 1 and self.selected_cell_domain_key.get() == DomainSelectionConstants.USER_SELECTION
        ):
            label = f"Selected cell index {selected_cells[0]}"
            show_clear = True
        elif len(selected_cells) > 1 and self.selected_cell_domain_key.get() == DomainSelectionConstants.USER_SELECTION:
            label = f"Selected {len(selected_cells)} cells"
            show_clear = True
        else:
            modifier = "cell" if len(selected_cells) == 1 else "cells"
            label = f"Selected cell domain {self.selected_cell_domain_key.get()} ({len(selected_cells)} {modifier})"
            show_clear = True
        children = [html.B(label, className="gr-breadcrumb-label")]

        if show_clear:
            children.append(
                html.Div(
                    [html.I(className="bi bi-x-circle")],
                    id="reset-selection-button",
                    n_clicks=0,
                    className="btn btn-link",
                    title="Clear selection",
                )
            )
        return html.Div(children)

    def _render_cell_selection_panes(self) -> Component:
        return [
            dbc.Col(self._render_cell_selection_title(panel_id="cell-selection-title-tree"), width=6),
            dbc.Col(self._render_cell_selection_title(panel_id="cell-selection-title-umap"), width=6),
        ]

    def _render_cell_selection_title(self, panel_id: str) -> Component:
        if self.selected_cl_name is not None:
            title = f"Selected cell class: {self.cl.cl_names_to_labels_map[self.selected_cl_name]}"
        else:
            title = DEFAULT_PANEL_TITLES[panel_id]

        return [html.Div(title, className="gr-header", id=panel_id)]

    def _render_closed_settings_pane(self) -> Component:
        return [
            html.Div(
                [
                    html.Label("Cell selection:", style={"margin-bottom": "5px"}),
                    self._render_domain_dropdown(),
                ],
                className="gr-form-item",
            ),
            html.Div(
                [
                    dbc.Label("Evidence threshold:", html_for="evidence-threshold"),
                    dcc.Slider(
                        id="evidence-threshold",
                        min=0,
                        max=1,
                        value=self.score_threshold.get(dirty_read=True),
                        marks={
                            0: "0",
                            0.25: "0.25",
                            0.5: "0.5",
                            0.75: "0.75",
                            1: "1",
                        },
                        tooltip={"placement": "bottom", "always_visible": True, "style": {"margin": "0 5px"}},
                    ),
                ],
                className="gr-form-item",
            ),
            html.Div(
                [
                    dbc.Label("Minimum cell fraction:", html_for="cell-fraction"),
                    dcc.Slider(
                        id="cell-fraction",
                        min=0,
                        max=1,
                        value=self.min_cell_fraction.get(dirty_read=True),
                        marks={
                            0: "0",
                            0.25: "0.25",
                            0.5: "0.5",
                            0.75: "0.75",
                            1: "1",
                        },
                        tooltip={"placement": "bottom", "always_visible": True, "style": {"margin": "0 5px"}},
                    ),
                ],
                className="gr-form-item",
            ),
            html.Div(
                [
                    dbc.Button(
                        "Cancel",
                        id="cancel-button",
                        title="Cancel the changes and close the settings pane",
                        n_clicks=0,
                    ),
                    dbc.Button(
                        "Update",
                        id="update-button",
                        title="Update the graphs based on the specified configuration",
                        n_clicks=0,
                    ),
                ],
                className="gr-settings-button-bar",
            ),
            html.A(
                html.Img(
                    src="assets/cellarium-powered-400px.png",
                ),
                href="https://cellarium.ai",
                className="gr-powered-by",
                target="_blank",
            ),
        ]

    def _render_domain_dropdown(self) -> Component:
        labels = [{"label": "None selected", "value": DomainSelectionConstants.NONE}]
        if len(self.selected_cells) > 0:
            labels.append({"label": "User selection", "value": DomainSelectionConstants.USER_SELECTION})

        if len(self.cell_domain_map.keys()) > 1:
            labels.append({"label": "________________", "value": None, "disabled": True})
            labels.append({"label": html.Span("Provided domains"), "value": None, "disabled": True})

            for k in list(self.cell_domain_map.keys())[1:]:
                labels.append({"label": k, "value": k})

        return dcc.Dropdown(
            id="domain-dropdown",
            options=labels,
            value=self.selected_cell_domain_key.get(),  # default to no selection
            className="gr-custom-dropdown",
            clearable=False,
        )

    def _get_effective_selected_cells(self) -> list:
        # User has chosen not to show any highlighted cells
        if self.selected_cell_domain_key.get() == DomainSelectionConstants.NONE:
            return []

        # User has chose to highlight explicitly selected cells
        if self.selected_cell_domain_key.get() == DomainSelectionConstants.USER_SELECTION:
            return self.selected_cells

        # User has chose to highlight pre-calculated domain cells
        if self.selected_cell_domain_key.get() is not None:
            return self.cell_domain_map[self.selected_cell_domain_key.get()]

    def _clear_cell_selection(self):
        self.selected_cells = []
        self.selected_cell_domain_key.reset()

    def _clear_cell_class_selection(self):
        self.selected_cl_name = None
        self.circular_tree_plot.update_selected_nodes(selected_cl_path=[])

    def _setup_callbacks(self) -> None:
        # Cell selection callbacks
        @self.app.callback(
            Output("circular-tree-plot", "figure", allow_duplicate=True),
            Output("umap-scatter-plot", "figure", allow_duplicate=True),
            Output("panel-titles", "children", allow_duplicate=True),
            Input("circular-tree-plot", "clickData"),
            prevent_initial_call=True,
        )
        def _update_umap_scatter_plot_based_on_circular_tree_plot(clickData):
            if clickData is None or "points" not in clickData:
                return (
                    self.circular_tree_plot.plotly_figure,
                    self._umap_scatter_plot_figure,
                    self._render_cell_selection_panes(),
                )

            point = clickData["points"][0]
            if "pointIndex" not in point:
                return (
                    self._umap_scatter_plot_figure,
                    self._umap_scatter_plot_figure,
                    self._render_cell_selection_panes(),
                )

            node_index = point["pointIndex"]
            selected_cl_name = self.circular_tree_plot.clade_index_to_cl_name_map.get(node_index)
            if selected_cl_name is not None:
                # Toggle selection if the selected cell class was clicked
                if selected_cl_name == self.selected_cl_name:
                    self._clear_cell_class_selection()
                else:
                    selected_cl_path = self.circular_tree_plot.get_clade_path_from_index(selected_cl_idx=node_index)
                    self.circular_tree_plot.update_selected_nodes(selected_cl_path)
                    self.selected_cl_name = selected_cl_name

            selected_cells_set = set(self._get_effective_selected_cells())

            if self.selected_cl_name is None:
                # Optimizing for lower code complexity over performance by always treating color and opacity as arrays in this case
                color = [self.umap_default_cell_color] * self.adata.n_obs
                opacity = [self.umap_default_opacity] * self.adata.n_obs
            else:
                scores = self._get_scores_for_cl_name(self.selected_cl_name)
                opacity = self._get_scatter_plot_opacity_from_scores(scores)
                color = sample_colorscale(self.circular_tree_plot.score_colorscale, scores)

            selected_cells_set = set(self._get_effective_selected_cells())
            # if no cells are selected but a cell class is, highlight all cells
            if len(selected_cells_set) == 0 and self.selected_cl_name is not None:
                selected_cells_set = set(self.cell_domain_map[self.ALL_CELLS_DOMAIN_KEY])

            for i_obs in range(self.adata.n_obs):
                if i_obs not in selected_cells_set:
                    color[i_obs] = self.umap_inactive_cell_color
                    opacity[i_obs] = self.umap_inactive_cell_opacity

            self._umap_scatter_plot_figure.update_traces(
                marker=dict(
                    color=color,
                    colorscale=self.circular_tree_plot.score_colorscale,
                    opacity=opacity,
                    cmin=0.0,
                    cmax=1.0,
                ),
                text=[f"{score:.5f}" for score in scores] if self.selected_cl_name is not None else None,
                hovertemplate=(
                    "<b>Evidence score: %{text}</b><extra></extra>" if self.selected_cl_name is not None else None
                ),
            )

            self._umap_scatter_plot_figure.update_layout(
                plot_bgcolor="white",
                margin=dict(l=0, r=25, t=50, b=0),
                # uirevision is needed to maintain pan/zoom state.  It must be updated to trigger a refresh
                uirevision=self._umap_scatter_plot_figure["layout"]["uirevision"],
            )
            return (
                self.circular_tree_plot.plotly_figure,
                self._umap_scatter_plot_figure,
                self._render_cell_selection_panes(),
            )

        @self.app.callback(
            Output("circular-tree-plot", "figure", allow_duplicate=True),
            Output("umap-scatter-plot", "figure", allow_duplicate=True),
            Output("selected-domain-label", "children", allow_duplicate=True),
            Output("panel-titles", "children", allow_duplicate=True),
            Input("umap-scatter-plot", "clickData"),
            State("umap-scatter-plot", "selectedData"),
            prevent_initial_call=True,
        )
        def _update_circular_tree_plot_based_on_umap_scatter_plot(clickData, selectedData):
            self._umap_scatter_plot_figure.update_selections()
            if clickData is None or "points" not in clickData:
                return (
                    self.circular_tree_plot.plotly_figure,
                    self._umap_scatter_plot_figure,
                    self._render_breadcrumb(),
                    self._render_cell_selection_panes(),
                )

            point = clickData["points"][0]

            if "pointIndex" not in point:
                return (
                    self.circular_tree_plot.plotly_figure,
                    self._umap_scatter_plot_figure,
                    self._render_breadcrumb(),
                    self._render_cell_selection_panes(),
                )

            node_index = point["pointIndex"]
            self.selected_cells = [node_index]
            self.selected_cell_domain_key.set(DomainSelectionConstants.USER_SELECTION).commit()
            self._clear_cell_class_selection()
            self._initialize_circular_tree_plot()
            self._initialize_umap_scatter_plot()

            return (
                self.circular_tree_plot.plotly_figure,
                self._umap_scatter_plot_figure,
                self._render_breadcrumb(),
                self._render_cell_selection_panes(),
            )

        @self.app.callback(
            Output("circular-tree-plot", "figure", allow_duplicate=True),
            Output("umap-scatter-plot", "figure", allow_duplicate=True),
            Output("umap-scatter-plot", "selectedData", allow_duplicate=True),
            Output("selected-domain-label", "children", allow_duplicate=True),
            Output("panel-titles", "children", allow_duplicate=True),
            Input("umap-scatter-plot", "selectedData"),
            prevent_initial_call=True,
        )
        def _update_circular_tree_plot_based_on_umap_scatter_plot_select(selectedData):
            # A selection event is firing on initialization. Ignore it by only accepting selectedData with a range field or lasso field
            if (
                selectedData is None
                or "points" not in selectedData
                or ("range" not in selectedData and "lassoPoints" not in selectedData)
            ):
                return (
                    self.circular_tree_plot.plotly_figure,
                    self._umap_scatter_plot_figure,
                    None,
                    self._render_breadcrumb(),
                    self._render_cell_selection_panes(),
                )

            points = selectedData["points"]

            node_indexes = [point["pointIndex"] for point in points]
            self.selected_cells = node_indexes
            self.selected_cell_domain_key.set(DomainSelectionConstants.USER_SELECTION).commit()
            self._clear_cell_class_selection()
            self._initialize_circular_tree_plot()
            self._initialize_umap_scatter_plot()
            return (
                self.circular_tree_plot.plotly_figure,
                self._umap_scatter_plot_figure,
                selectedData,
                self._render_breadcrumb(),
                self._render_cell_selection_panes(),
            )

        @self.app.callback(
            Output("circular-tree-plot", "figure", allow_duplicate=True),
            Output("umap-scatter-plot", "figure", allow_duplicate=True),
            Output("selected-domain-label", "children", allow_duplicate=True),
            Output("settings-pane", "children", allow_duplicate=True),
            Output("panel-titles", "children", allow_duplicate=True),
            Input("reset-selection-button", "n_clicks"),
            prevent_initial_call=True,
        )
        def _reset_selection(n_clicks):
            if n_clicks != 0:
                self._clear_cell_selection()
                self._clear_cell_class_selection()

                # update the figures
                self._initialize_circular_tree_plot()
                self._initialize_umap_scatter_plot()

            return (
                self.circular_tree_plot.plotly_figure,
                self._umap_scatter_plot_figure,
                self._render_breadcrumb(),
                self._render_closed_settings_pane(),
                self._render_cell_selection_panes(),
            )

        # Settings callbacks
        @self.app.callback(
            Output("circular-tree-plot", "figure", allow_duplicate=True),
            Output("umap-scatter-plot", "figure", allow_duplicate=True),
            Output("selected-domain-label", "children", allow_duplicate=True),
            Output("settings-pane", "children", allow_duplicate=True),
            Output("panel-titles", "children", allow_duplicate=True),
            Output("settings-pane", "is_open", allow_duplicate=True),
            Input("update-button", "n_clicks"),
            prevent_initial_call=True,
        )
        def _save_settings(n_clicks):
            if n_clicks > 0:
                # If a domain selection was changed and set to None, clear all selections
                if (
                    self.selected_cell_domain_key.is_dirty()
                    and self.selected_cell_domain_key.get(dirty_read=True) is DomainSelectionConstants.NONE
                ):
                    self._clear_cell_selection()

                self.selected_cell_domain_key.commit()
                self.score_threshold.commit()
                self.min_cell_fraction.commit()

                self._clear_cell_class_selection()

                # update the figures
                self._initialize_circular_tree_plot()
                self._initialize_umap_scatter_plot()

            return (
                self.circular_tree_plot.plotly_figure,
                self._umap_scatter_plot_figure,
                self._render_breadcrumb(),
                self._render_closed_settings_pane(),
                self._render_cell_selection_panes(),
                False,
            )

        @self.app.callback(
            Output("settings-pane", "children", allow_duplicate=True),
            Output("settings-pane", "is_open", allow_duplicate=True),
            Input("cancel-button", "n_clicks"),
            prevent_initial_call=True,
        )
        def _cancel_settings(n_clicks):
            self.selected_cell_domain_key.rollback()
            self.score_threshold.rollback()
            self.min_cell_fraction.rollback()

            return self._render_closed_settings_pane(), False

        @self.app.callback(
            Output("settings-pane", "is_open", allow_duplicate=True),
            Input("settings-button", "n_clicks"),
            [State("settings-pane", "is_open")],
            prevent_initial_call=True,
        )
        def _toggle_settings(n_clicks, is_open):
            if n_clicks:
                return not is_open
            return is_open

        @self.app.callback(
            Output("no-action", "children", allow_duplicate=True),
            Input("domain-dropdown", "value"),
            prevent_initial_call=True,
        )
        def _update_domain(domain):
            # set the domain
            self.selected_cell_domain_key.set(domain)

        @self.app.callback(
            Output("no-action", "children", allow_duplicate=True),
            Input("evidence-threshold", "value"),
            prevent_initial_call=True,
        )
        def _update_evidence_threshold(input_value):
            try:
                self.score_threshold.set(float(input_value))
            except ValueError:
                pass
            return input_value

        @self.app.callback(
            Output("no-action", "children", allow_duplicate=True),
            Input("cell-fraction", "value"),
            prevent_initial_call=True,
        )
        def _update_cell_fraction(input_value):
            try:
                self.min_cell_fraction.set(float(input_value))
            except ValueError:
                pass
            return input_value



import platform
import subprocess
import typing as t

from typing_extensions import Self


def find_and_kill_process(port: int, verbose: bool = False):
    """
    Find and kill processes using a specific port.

    :param port: The port number to search for and kill processes.
    :type port: int

    :raises OSError: If the operating system is not supported.

    :return: None
    """

    system = platform.system()

    if system == "Linux" or system == "Darwin":  # macOS is Darwin
        find_command = f"lsof -i :{port}"
        kill_command = "kill -9"

    elif system == "Windows":
        find_command = f"netstat -ano | findstr :{port}"
        kill_command = "taskkill /PID"

    else:
        raise OSError(f"Unsupported operating system: {system}")

    try:
        # Find the process using the port
        result = subprocess.run(find_command, shell=True, check=True, capture_output=True, text=True)
        output = result.stdout

        if system == "Linux" or system == "Darwin":
            lines = output.strip().split("\n")
            for line in lines[1:]:
                pid = int(line.split()[1])
                subprocess.run(f"{kill_command} {pid}", shell=True)
                if verbose:
                    print(f"Killed process with PID: {pid}")

        elif system == "Windows":
            lines = output.strip().split("\n")
            for line in lines:
                pid = int(line.split()[-1])
                subprocess.run(f"{kill_command} {pid} /F", shell=True)
                if verbose:
                    print(f"Killed process with PID: {pid}")

        if verbose:
            print(f"All processes using port {port} have been killed.")

    except subprocess.CalledProcessError as e:
        if verbose:
            print(f"No process found using port {port}. {e}")


T = t.TypeVar("T")


class ConfigValue(t.Generic[T]):
    """
    Class to store a configuration value and its original value.  Useful for tracking uncommitted changes.

    Note: there is no typechecking to ensure that the original value and the new value are of the same type.
    """

    def __init__(self, value: T):
        self.__value__ = value
        self.__original_value__ = value
        self.__dirty_value__ = None
        self.__is_dirty__ = False

    def set(self, value: T) -> Self:
        self.__dirty_value__ = value
        self.__is_dirty__ = True
        return self

    def rollback(self) -> Self:
        self.__dirty_value__ = None
        self.__is_dirty__ = False
        return self

    def commit(self) -> Self:
        if self.__is_dirty__:
            self.__value__ = self.__dirty_value__
        self.__is_dirty__ = False
        return self

    def get(self, dirty_read: bool = False) -> T:
        if dirty_read and self.__is_dirty__:
            return self.__dirty_value__
        return self.__value__

    def is_dirty(self) -> bool:
        return self.__is_dirty__

    def reset(self) -> Self:
        self.__value__ = self.__original_value__
        self.__dirty_value__ = None
        self.__is_dirty__ = False
        return self



import logging
import typing as t
from functools import lru_cache
from logging import log

import networkx as nx
import owlready2
from scipy import sparse as sp

from cellarium.cas._io import suppress_stderr

# Used in CZ CELLxGENE schema v3:
# https://github.com/chanzuckerberg/single-cell-curation/blob/main/schema/3.0.0/schema.md
DEFAULT_CL_OWL_PATH = "https://github.com/obophenotype/cell-ontology/raw/v2022-09-15/cl.owl"

# only keep nodes with the following prefix when parsing CL ontology
CL_PREFIX = "CL_"

# the 'cell' node
CL_CELL_ROOT_NODE = "CL_0000000"

# the 'eukaryotic cell' node
CL_EUKARYOTIC_CELL_ROOT_NODE = "CL_0000255"


class CellOntologyCache:
    """
    A class representing a cache for the Cell Ontology (CL).

    Args:
        cl_owl_path (str): The path to the Cell Ontology OWL file. Defaults to DEFAULT_CL_OWL_PATH.

    Attributes:
        cl_classes (list): A list of Cell Ontology classes with a singleton label.
        cl_names (list): A list of names of Cell Ontology classes.
        cl_labels (list): A list of labels of Cell Ontology classes.
        cl_classes_set (set): A set of Cell Ontology classes.
        cl_names_to_labels_map (dict): A mapping of Cell Ontology class names to labels.
        cl_labels_to_names_map (dict): A mapping of Cell Ontology class labels to names.
        cl_names_to_idx_map (dict): A mapping of Cell Ontology class names to indices.
        cl_idx_to_names_map (dict): A mapping of indices to Cell Ontology class names.
        cl_graph (nx.DiGraph): A networkx graph representing the Cell Ontology.

    """

    def __init__(self, cl_owl_path: str = DEFAULT_CL_OWL_PATH):
        """
        Initialize the CellOntologyCache object.

        Loads the Cell Ontology from the specified OWL file, filters out classes with a singleton label,
        and builds a networkx graph representing the Cell Ontology.

        :param cl_owl_path: The path to the Cell Ontology OWL file. Defaults to DEFAULT_CL_OWL_PATH.
        :type cl_owl_path: str

        """

        with suppress_stderr():
            log(logging.INFO, f"Loading cell ontology OWL from:\n{cl_owl_path}")
            cl = owlready2.get_ontology(cl_owl_path).load()

        # only keep CL classes with a singleton label
        cl_classes = list(
            _class for _class in cl.classes() if _class.name.startswith(CL_PREFIX) and len(_class.label) == 1
        )

        cl_names = [_class.name for _class in cl_classes]
        cl_labels = [_class.label[0] for _class in cl_classes]
        assert len(set(cl_names)) == len(cl_classes)
        assert len(set(cl_labels)) == len(cl_classes)

        cl_classes_set = set(cl_classes)
        cl_names_to_labels_map = {cl_name: cl_label for cl_name, cl_label in zip(cl_names, cl_labels)}
        cl_labels_to_names_map = {cl_label: cl_name for cl_name, cl_label in zip(cl_names, cl_labels)}
        cl_names_to_idx_map = {cl_name: idx for idx, cl_name in enumerate(cl_names)}
        cl_idx_to_names_map = {idx: cl_name for idx, cl_name in enumerate(cl_names)}

        # build a networkx graph from CL
        cl_graph = nx.DiGraph(name="CL graph")

        for cl_class in cl_classes:
            cl_graph.add_node(cl_class.name)

        for self_cl_class in cl_classes:
            for parent_cl_class in cl.get_parents_of(self_cl_class):
                if parent_cl_class not in cl_classes_set:
                    continue
                cl_graph.add_edge(parent_cl_class.name, self_cl_class.name)
            for child_cl_class in cl.get_children_of(self_cl_class):
                if child_cl_class not in cl_classes_set:
                    continue
                cl_graph.add_edge(self_cl_class.name, child_cl_class.name)

        self.cl_classes = cl_classes
        self.cl_names = cl_names
        self.cl_labels = cl_labels
        self.cl_classes_set = cl_classes_set
        self.cl_names_to_labels_map = cl_names_to_labels_map
        self.cl_labels_to_names_map = cl_labels_to_names_map
        self.cl_names_to_idx_map = cl_names_to_idx_map
        self.cl_idx_to_names_map = cl_idx_to_names_map
        self.cl_graph = cl_graph

    @property
    @lru_cache(maxsize=None)
    def cl_ancestors_csr_matrix(self) -> sp.csr_matrix:
        """Returns a sparse matrix representation of ancestors.

        .. note:
            The matrix element (i, j) = 1 iff j is an ancetor of i.
        """
        n_nodes = len(self.cl_graph.nodes)

        row = []
        col = []
        data = []

        for cl_name in self.cl_names:
            self_idx = self.cl_names_to_idx_map[cl_name]
            for cl_ancestor_name in nx.ancestors(self.cl_graph, cl_name):
                ancestor_idx = self.cl_names_to_idx_map[cl_ancestor_name]
                row.append(self_idx)
                col.append(ancestor_idx)
                data.append(1)

        cl_ancestors_csr_matrix = sp.csr_matrix((data, (row, col)), shape=(n_nodes, n_nodes))
        return cl_ancestors_csr_matrix

    @lru_cache(maxsize=None)
    def get_longest_path_lengths_from_target(self, target: str) -> t.Dict[str, float]:
        # Perform a topological sort of the graph
        topo_order = list(nx.topological_sort(self.cl_graph))

        # Initialize distances with -infinity for all nodes except the target
        distances = {node: float("-inf") for node in self.cl_graph.nodes()}
        distances[target] = 0

        # Process nodes in topological order
        for node in topo_order:
            if distances[node] != float("-inf"):  # Only process reachable nodes
                for neighbor in self.cl_graph.successors(node):
                    if distances[neighbor] < distances[node] + 1:
                        distances[neighbor] = distances[node] + 1

        return distances



# classes
from .cell_type_summary_statistics import (
    get_interpolated_cell_type_colors,
    reduce_cas_cell_type_summary_statistics_response_by_majority_vote_per_cluster,
    reduce_cas_cell_type_summary_statistics_response_by_min_distance,
    reduce_cas_cell_type_summary_statistics_response_by_wnn,
    reduce_cas_query_result_by_majority_vote,
)
from .common import get_obs_indices_for_cluster

# constants
# methods
from .ontology_aware import (
    CAS_CL_LABEL_PROPERTY_REF,
    CAS_CL_SCORES_ANNDATA_OBSM_KEY,
    CAS_FRACTION_PROPERTY_REF,
    CAS_METADATA_ANNDATA_UNS_KEY,
    CAS_SCORE_PROPERTY_REF,
    AggregatedCellOntologyScores,
    CellOntologyScoresAggregationDomain,
    CellOntologyScoresAggregationOp,
    compute_most_granular_top_k_calls_cluster,
    compute_most_granular_top_k_calls_single,
    convert_aggregated_cell_ontology_scores_to_rooted_tree,
    convert_cas_ontology_aware_response_to_score_matrix,
    generate_phyloxml_from_scored_cell_ontology_tree,
    get_aggregated_cas_ontology_aware_scores,
    get_most_granular_top_k_calls,
    insert_cas_ontology_aware_response_into_adata,
)

__all__ = [
    "AggregatedCellOntologyScores",
    "CellOntologyScoresAggregationOp",
    "CellOntologyScoresAggregationDomain",
    "convert_cas_ontology_aware_response_to_score_matrix",
    "insert_cas_ontology_aware_response_into_adata",
    "get_aggregated_cas_ontology_aware_scores",
    "convert_aggregated_cell_ontology_scores_to_rooted_tree",
    "generate_phyloxml_from_scored_cell_ontology_tree",
    "get_obs_indices_for_cluster",
    # ontology aware
    "get_most_granular_top_k_calls",
    "compute_most_granular_top_k_calls_single",
    "compute_most_granular_top_k_calls_cluster",
    # cell type summary statistics
    "reduce_cas_query_result_by_majority_vote",
    "reduce_cas_cell_type_summary_statistics_response_by_min_distance",
    "reduce_cas_cell_type_summary_statistics_response_by_majority_vote_per_cluster",
    "reduce_cas_cell_type_summary_statistics_response_by_wnn",
    "get_interpolated_cell_type_colors",
    # constants
    "CAS_CL_SCORES_ANNDATA_OBSM_KEY",
    "CAS_METADATA_ANNDATA_UNS_KEY",
    "CAS_SCORE_PROPERTY_REF",
    "CAS_FRACTION_PROPERTY_REF",
    "CAS_CL_LABEL_PROPERTY_REF",
]



NUM_ATTEMPTS_PER_CHUNK_DEFAULT = 7
MAX_NUM_REQUESTS_AT_A_TIME = 8
START_RETRY_DELAY = 5
MAX_RETRY_DELAY = 32
AIOHTTP_TOTAL_TIMEOUT_SECONDS = 750
AIOHTTP_READ_TIMEOUT_SECONDS = 730
MAX_CHUNK_SIZE_SEARCH_METHOD = 500


def is_interactive_environment() -> bool:
    """
    Check if the current environment is interactive (e.g. Jupyter notebook, IPython terminal)

    :return: True if the current environment is interactive, False otherwise
    """
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:
            # Running in a non-interactive environment.
            return False
        else:
            if "terminal" in get_ipython().config["IPKernelApp"]["connection_file"]:
                # Running in an IPython terminal
                return True
            else:
                # Running in a Jupyter environment.
                return True
    except (ImportError, AttributeError):
        # Running in a non-interactive environment.
        return False



import asyncio
import json
import ssl
import typing as t
from contextlib import ContextDecorator
from contextvars import ContextVar
from uuid import UUID, uuid4

import aiohttp
import certifi
import nest_asyncio
import requests
from aiohttp import client_exceptions

from cellarium.cas import constants, endpoints, exceptions, settings

if settings.is_interactive_environment():
    print("Running in an interactive environment, applying nest_asyncio")
    nest_asyncio.apply()

# Context variable to track the action id for the current context
client_action_id = ContextVar("action_id", default=None)


class action_context_manager(ContextDecorator):
    """
    Handle the lifecycle of an action id for the current context.

    Add the `@action_context_manager` decorator around methods in the client methods that should
    have a common action id.
    """

    def __enter__(self):
        self.t = client_action_id.set(uuid4())
        return self

    def __exit__(self, *exc):
        client_action_id.reset(self.t)


class _BaseService:
    """
    Base class for communicating with a Cellarium Cloud API service
    It leverages async request library `aiohttp` to asynchronously execute HTTP requests.

    :param api_token: A token that could be authenticated by Cellarium Cloud Backend API service
    :param api_url: URL of the Cellarium Cloud Backend API service
    :param client_session_id: A unique identifier for the current client session
    """

    def __init__(
        self,
        api_token: str,
        api_url: str = settings.CELLARIUM_CLOUD_BACKEND_URL,
        client_session_id: UUID = None,
        *args,
        **kwargs,
    ):
        self.api_token = api_token
        self.api_url = api_url
        self.client_session_id = client_session_id
        super().__init__(*args, **kwargs)

    def _get_endpoint_url(self, endpoint: str) -> str:
        """
        Configure a specific method endpoint from backend url and endpoint

        :param endpoint: Endpoint string without a leading slash
        :return: Full url with backend domains/subdomains and endpoint joined
        """
        return f"{self.api_url}/{endpoint}"

    def _get_headers(self) -> t.Dict[str, str]:
        """
        Get the headers to include in the request

        :return: Headers dictionary
        """
        headers = {constants.Headers.authorization: f"Bearer {self.api_token}"}
        if self.client_session_id:
            headers[constants.Headers.client_session_id] = str(self.client_session_id)
        action_id = client_action_id.get()
        if action_id:
            headers[constants.Headers.client_action_id] = str(action_id)
        return headers

    @staticmethod
    def raise_response_exception(status_code: int, detail: str) -> None:
        """
        Raise an exception based on the status code returned by the server, including the detail message

        :param status_code: HTTP status code
        :param detail: Detail message returned by the server
        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError
        """
        message = f"Server returned status code {status_code}, Detail: {detail}"
        if status_code == constants.HTTP.STATUS_401_UNAUTHORIZED:
            raise exceptions.HTTPError401(message)
        elif status_code == constants.HTTP.STATUS_403_FORBIDDEN:
            raise exceptions.HTTPError403(message)
        elif status_code == constants.HTTP.STATUS_NOT_FOUND:
            raise exceptions.HTTPError404(message)

        elif (
            constants.HTTP.STATUS_500_INTERNAL_SERVER_ERROR
            <= status_code
            <= constants.HTTP.STATUS_511_NETWORK_AUTHENTICATION_REQUIRED
        ):
            raise exceptions.HTTPError5XX(message)
        else:
            raise exceptions.HTTPError(message)

    def __validate_requests_response(self, response: requests.Response) -> None:
        """
        Validate requests response and raise an exception if response status code is not 200

        :param response: Response object

        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError
        """
        status_code = response.status_code
        if not (constants.HTTP.STATUS_200_OK <= status_code <= constants.HTTP.STATUS_226_IM_USED):
            # When response status code is not 2XX
            try:
                response_detail = response.json()["detail"]
            except (json.decoder.JSONDecodeError, KeyError):
                response_detail = response.text

            self.raise_response_exception(status_code=status_code, detail=response_detail)

    def get(self, endpoint: str) -> requests.Response:
        """
        Make a GET request to backend service

        :param endpoint: Endpoint string without a leading slash

        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError

        :return: Response object
        """
        url = self._get_endpoint_url(endpoint)
        headers = self._get_headers()
        response = requests.get(url=url, headers=headers)
        self.__validate_requests_response(response=response)
        return response

    def post(self, endpoint: str, data: t.Optional[t.Dict] = None) -> requests.Response:
        url = self._get_endpoint_url(endpoint)
        headers = self._get_headers()
        response = requests.post(url=url, headers=headers, json=data)
        self.__validate_requests_response(response=response)
        return response

    def get_json(self, endpoint: str) -> t.Union[t.Dict, t.List]:
        """
        Make a GET request to backend service and return JSON response

        :param endpoint: Endpoint string without a leading slash

        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError

        :return: JSON response
        """
        return self.get(endpoint=endpoint).json()

    def post_json(self, endpoint: str, data: t.Optional[t.Dict] = None) -> t.Union[t.Dict, t.List]:
        """
        Make a POST request to backend service and return JSON response

        :param endpoint: Endpoint string without a leading slash
        :param data: Dictionary to include to HTTP POST request body

        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError

        :return: JSON response
        """
        return self.post(endpoint=endpoint, data=data).json()

    async def _aiohttp_async_post(
        self, url: str, form_data: t.Optional[aiohttp.FormData] = None, headers: t.Optional[t.Dict[str, t.Any]] = None
    ):
        """
        Make an async POST request to backend service with timeout and SSL context, handle exceptions and return JSON.

        Create custom SSL context with `certifi` package to avoid SSL errors. Use :class:`aiohttp.ClientTimeout` to set
        timeout for the request. Handle exceptions and raise custom exceptions.

        :param url: Endpoint URL
        :param form_data: :class:`aiohttp.FormData` object
        :param headers: Headers to include in the request

        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError, HTTPClientError

        :return: JSON response
        """
        timeout = aiohttp.ClientTimeout(
            total=settings.AIOHTTP_TOTAL_TIMEOUT_SECONDS, sock_read=settings.AIOHTTP_READ_TIMEOUT_SECONDS
        )
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = aiohttp.TCPConnector(ssl=ssl_context)

        async with aiohttp.ClientSession(connector=connector, timeout=timeout) as session:
            try:
                async with session.post(url, data=form_data, headers=headers) as response:
                    status_code = response.status
                    if status_code < 200 or status_code >= 300:
                        try:
                            response_body = await response.json()
                            response_detail = response_body["detail"]
                        except (json.decoder.JSONDecodeError, client_exceptions.ClientResponseError):
                            response_detail = await response.text()
                        except KeyError:
                            print("Response body doesn't have a 'detail' key, returning full response body")
                            response_detail = str(await response.json())

                        self.raise_response_exception(status_code=status_code, detail=response_detail)
                    try:
                        return await response.json()
                    except client_exceptions.ClientPayloadError as e:
                        raise exceptions.HTTPClientError(f"Failed to parse response body: {e.__class__.__name__}: {e}")

            except (client_exceptions.ServerTimeoutError, asyncio.TimeoutError) as e:
                raise exceptions.HTTPClientError(f"Client Timeout Error: {e}")
            except (client_exceptions.ClientConnectionError, client_exceptions.ClientResponseError) as e:
                raise exceptions.HTTPClientError(f"Client Connection Error: {e.__class__.__name__}: {e}")
            except client_exceptions.ClientError as e:
                raise exceptions.HTTPClientError(f"Unknown Error: {e.__class__.__name__}: {e}")

    async def async_post(
        self,
        endpoint: str,
        file: t.ByteString,
        data: t.Optional[t.Dict] = None,
        headers: t.Optional[t.Dict] = None,
    ) -> t.Union[t.List, t.Dict]:
        """
        Make an async multiform POST request to backend service. Include adata file as a byte file in the request body,
        and authorization token in the headers.

        For details about custom settings for SSL context and timeout, refer to :meth:`_aiohttp_async_post` method.

        :param endpoint: Endpoint string without a leading slash
        :param file: Byte file to attach to request
        :param data: Dictionary to include to HTTP POST request body
        :param headers: Dictionary to include to HTTP POST request Headers

        :raises: HTTPError401, HTTPError403, HTTPError500, HTTPBaseError, HTTPClientError

        :return: JSON response
        """
        url = self._get_endpoint_url(endpoint=endpoint)
        _data = data if data is not None else {}
        _headers = self._get_headers()

        if headers is not None:
            _headers.update(headers)

        form_data = aiohttp.FormData()
        form_data.add_field("file", file, filename="adata.h5ad")

        for key, value in _data.items():
            form_data.add_field(key, value)

        return await self._aiohttp_async_post(url=url, form_data=form_data, headers=_headers)


class CASAPIService(_BaseService):
    """
    Class with all the API methods of Cellarium Cloud CAS infrastructure.
    """

    def validate_token(self) -> t.Dict[str, str]:
        """
        Validate user given API token.
        Would raise 401 Unauthorized if token is invalid.

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/validate_token_api_cellarium_general_validate_token_get

        :return: Dictionary with basic user information
        """
        return self.get_json(endpoint=endpoints.VALIDATE_TOKEN)

    def get_application_info(self) -> t.Dict[str, str]:
        """
        Retrieve General Application Info. This Includes default schema, version, model information, etc.

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/application_info_api_cellarium_general_application_info_get

        :return: Dictionary with application info
        """
        return self.get_json(endpoint=endpoints.APPLICATION_INFO)

    def feedback_opt_out(self) -> t.Dict[str, str]:
        """
        Opt out of future feedback requests.

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/feedback/opt-out
        :return: Dictionary with basic user information
        """
        return self.post_json(endpoint=endpoints.FEEDBACK_OPT_OUT)

    def get_feedback_answer_link(self) -> str:
        """
        Retrieve a link to answer feedback questions

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/feedback/answer
        :return: Link to answer feedback questions
        """
        return self._get_endpoint_url(
            endpoints.FEEDBACK_ANSWER.format(
                client_session_id=self.client_session_id, client_action_id=client_action_id.get()
            )
        )

    def validate_version(self, version_str: str) -> t.Dict[str, t.Any]:
        """
        Validate client version with the server to see if it is compatible.
        Would raise 400 Bad Request if version is not compatible.

        :return: Void
        """
        return self.post_json(endpoint=endpoints.VALIDATE_VERSION, data={"client_version": version_str})

    def get_feature_schemas(self) -> t.List[str]:
        """
        Retrieve a list of feature schemas that exist in Cellarium Cloud CAS

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/get_feature_schemas_api_cellarium_general_feature_schemas_get

        :return: List of feature schema names
        """
        return [x["schema_name"] for x in self.get_json(endpoint=endpoints.GET_FEATURE_SCHEMAS)]

    def get_feature_schema_by(self, name: str) -> t.List[str]:
        """
        Retrieve feature schema by name

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/get_feature_schema_by_api_cellarium_general_feature_schema__schema_name__get

        :param name: Name of feature schema

        :return: List of feature ids
        """
        return self.get_json(endpoint=endpoints.GET_SCHEMA_BY_NAME.format(schema_name=name))

    def get_model_list(self) -> t.List[t.Dict[str, t.Any]]:
        """
        Retrieve list of all models that are in CAS

        Refer to API Docs:
        {api_url}/api/docs#/cellarium-general/get_model_list_api_cellarium_general_list_models_get

        :return: List of models
        """
        return self.get_json(endpoint=endpoints.LIST_MODELS)

    def get_user_quota(self) -> t.Dict[str, t.Any]:
        """
        Retrieve user quota information

        Refer to API Docs:
        {BACKEND_URL}/api/docs

        :return: User quota information
        """
        return self.get_json(endpoint=endpoints.GET_USER_QUOTA)

    def query_cells_by_ids(
        self, model_name: str, cell_ids: t.List[int], metadata_feature_names: t.List[str]
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Retrieve cells by their ids from Cellarium Cloud database.

        Refer to API Docs:
        {api_url}/api/docs#/cell-analysis/get_cells_by_ids_api_cellarium_cas_query_cells_by_ids_post

        :param model_name: Name of the model to use. Model name is required to locate the correct database.
        :param cell_ids: List of cell ids from Cellarium Cloud database to query by.
        :param metadata_feature_names: List of metadata feature names to include in the response.

        :return: List of cells with metadata.
        """
        request_data = {
            "model_name": model_name,
            "cas_cell_ids": cell_ids,
            "metadata_feature_names": metadata_feature_names,
        }
        return self.post_json(endpoint=endpoints.QUERY_CELLS_BY_IDS, data=request_data)

    async def async_annotate_cell_type_summary_statistics_strategy(
        self, adata_bytes: t.ByteString, model_name: str, include_extended_output: bool
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Request Cellarium Cloud infrastructure to annotate an input anndata file using cell type count statistics

        Refer to API Docs: {api_url}/api/docs#/cell-analysis/annotate_api_cellarium_cas_annotate_post

        :param adata_bytes: Validated anndata file
        :param model_name: Name of the model to use.
        :param include_extended_output: Whether to include dev metadata in the response.

        :return: A list of dictionaries with annotations.
        """
        request_data = {
            "model_name": model_name,
            "include_extended_output": str(include_extended_output),
        }
        return await self.async_post(
            endpoints.ANNOTATE_CELL_TYPE_SUMMARY_STATS_STRATEGY, file=adata_bytes, data=request_data
        )

    async def async_annotate_cell_type_ontology_aware_strategy_anndata(
        self, adata_bytes: t.ByteString, model_name: str, prune_threshold: float, weighting_prefactor: float
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Request Cellarium Cloud infrastructure to annotate an input anndata file using ontology-aware strategy

        Refer to API Docs:
        {api_url}/api/docs#/cell-operations/annotate_ontology_aware_strategy_api_cellarium_cell_operations_annotate_ontology_aware_strategy_post

        :param adata_bytes: Validated anndata file
        :param model_name: Name of the model to use.
        :param prune_threshold: Whether to normalize consensus result.
        :param weighting_prefactor: Weighting prefactor for the ontology-aware strategy.

        :return: A list of dictionaries with annotations.
        """
        request_data = {
            "model_name": model_name,
            "prune_threshold": str(prune_threshold),
            "weighting_prefactor": str(weighting_prefactor),
        }
        return await self.async_post(
            endpoints.ANNOTATE_CELL_TYPE_ONTOLOGY_AWARE_STRATEGY, file=adata_bytes, data=request_data
        )

    async def async_nearest_neighbor_search(
        self, adata_bytes: t.ByteString, model_name: str
    ) -> t.List[t.Dict[str, t.Any]]:
        """
        Request Cellarium Cloud infrastructure to search for nearest neighbors in an input anndata file

        Refer tp API Docs:
        {api_url}/api/docs#/cell-analysis/nearest_neighbor_search_api_cellarium_cas_nearest_neighbor_search_post

        :param adata_bytes: Validated anndata file
        :param model_name: Name of the model to use.

        :return: A list of dictionaries with annotations (query_id and cas_cell_id, distance).
        """
        request_data = {
            "model_name": model_name,
        }
        return await self.async_post(endpoints.NEAREST_NEIGHBOR_SEARCH, file=adata_bytes, data=request_data)



import typing as t

from cellarium.cas import exceptions

if t.TYPE_CHECKING:
    import anndata


def validate(
    adata: "anndata.AnnData",
    cas_feature_schema_list: t.List,
    feature_ids_column_name: str,
) -> None:
    """
    Validate input `anndata.AnnData` instance in concordance with feature schema `cas_feature_schema_list`.
    Raise `exceptions.DataValidationError` with number of missing features and number of extra features
    that were present in dataset.

    :param adata: Instance to validate
    :param cas_feature_schema_list: List of features to be validated with
    :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. Default `index`.
    """
    assert feature_ids_column_name == "index" or feature_ids_column_name in adata.var.columns.values, (
        "`feature_ids_column_name` should have a value of either 'index' "
        "or be present as a column in the `adata.var` object."
    )

    if feature_ids_column_name == "index":
        adata_feature_schema_list = adata.var.index.tolist()
    else:
        adata_feature_schema_list = adata.var[feature_ids_column_name].values.tolist()

    cas_feature_schema_set = set(cas_feature_schema_list)
    adata_feature_schema_set = set(adata_feature_schema_list)

    if adata_feature_schema_list == cas_feature_schema_list:
        return

    missing_features = len(cas_feature_schema_set - adata_feature_schema_set)
    extra_features = len(adata_feature_schema_set - cas_feature_schema_set)

    raise exceptions.DataValidationError(missing_features=missing_features, extra_features=extra_features)



from __future__ import annotations

import typing as t
from collections import defaultdict
from operator import itemgetter

import numpy as np
from anndata import AnnData
from tqdm.notebook import tqdm

ALLOWED_WNN_STRATEGIES = {"connectivities"}


def reduce_cas_query_result_by_majority_vote(
    adata: AnnData,
    cas_cell_type_summary_statistics_response: list,
    output_cell_type_obs_column: str = "cas_cell_type",
    output_cell_type_confidence_score_obs_column: str = "cas_cell_type_confidence_score",
):
    """Reduce the CAS cell type summary statistics response by majority vote.

    This function takes the result of a CAS cell type summary statistics response (the output of
    :class:`cellarium.cas.client.annotate_matrix_cell_type_summary_statistics_strategy`) and reduces
    it by performing a majority vote to determine the most likely cell type for each query. It also
    calculates a confidence score based on the ratio of the best cell count to the total cell count.

    :param adata: An Annotated Data object containing the query result.
    :type adata: anndata.AnnData
    :param cas_cell_type_summary_statistics_response: The summary statistics response from the CAS query.
    :type cas_cell_type_summary_statistics_response: list
    :param output_cell_type_obs_column: The name of the observation column to store the predicted cell types.
    :type output_cell_type_obs_column: str, optional
    :param output_cell_type_confidence_score_obs_column: The name of the observation column to store the
      confidence scores.
    :type output_cell_type_confidence_score_obs_column: str, optional
    """

    majority_vote_cell_type_list = []
    majority_vote_confidence_score_list = []
    for single_cell_query in cas_cell_type_summary_statistics_response:
        total_cell_count = 0
        best_cell_count = -1
        best_cell_type = None
        for match in single_cell_query["matches"]:
            total_cell_count += match["cell_count"]
            if match["cell_count"] > best_cell_count:
                best_cell_count = match["cell_count"]
                best_cell_type = match["cell_type"]
        majority_vote_cell_type_list.append(best_cell_type)
        majority_vote_confidence_score_list.append(best_cell_count / total_cell_count)

    adata.obs[output_cell_type_obs_column] = majority_vote_cell_type_list
    adata.obs[output_cell_type_confidence_score_obs_column] = majority_vote_confidence_score_list


def reduce_cas_cell_type_summary_statistics_response_by_min_distance(
    adata: AnnData,
    cas_cell_type_summary_statistics_response: list,
    output_cell_type_obs_column: str = "cas_cell_type",
    output_cell_type_min_distance_obs_column: str = "cas_cell_type_min_distance",
):
    """
    Reduce the CAS query result by selecting the cell type with the minimum distance for each query.

    :param adata: Annotated data object.
    :type adata: AnnData
    :param cas_cell_type_summary_statistics_response: CAS response.
    :type cas_cell_type_summary_statistics_response: list
    :param output_cell_type_obs_column: Name of the output observation column for cell types.
      (default: "cas_cell_type")
    :type output_cell_type_obs_column: str
    :param output_cell_type_min_distance_obs_column: Name of the output observation column for minimum distances.
      (default: "cas_cell_type_min_distance")
    :type output_cell_type_min_distance_obs_column: str
    :return: None
    """

    min_distance_cell_type_list = []
    min_distance_list = []
    for single_cell_query in cas_cell_type_summary_statistics_response:
        min_distance = np.inf
        best_cell_type = None
        for match in single_cell_query["matches"]:
            if match["min_distance"] < min_distance:
                min_distance = match["min_distance"]
                best_cell_type = match["cell_type"]
        min_distance_cell_type_list.append(best_cell_type)
        min_distance_list.append(min_distance)

    adata.obs[output_cell_type_obs_column] = min_distance_cell_type_list
    adata.obs[output_cell_type_min_distance_obs_column] = min_distance_list


def reduce_cas_cell_type_summary_statistics_response_by_majority_vote_per_cluster(
    adata: AnnData,
    cas_cell_type_summary_statistics_response: dict,
    cluster_key: str = "leiden",
    output_cell_type_obs_column: str = "cas_per_cluster_cell_type",
    output_cell_type_confidence_score_obs_column: str = "cas_per_cluster_cell_type_confidence_score",
    ignore_set: t.Optional[set] = None,
) -> dict:
    """
    Reduce the CAS query result by performing a majority vote per cluster.

    This function takes the result of a CAS query and reduces it by performing a majority vote per cluster.
    It calculates the most likely cell type for each cluster based on the majority vote and assigns it to the
    corresponding cells in the query result. It also calculates a confidence score based on the ratio of the
    best cell count to the total cell count.

    :param adata: An Annotated Data object containing the query result.
    :type adata: anndata.AnnData
    :param cas_query_res: The CAS query result dictionary.
    :type cas_query_res: dict
    :param cluster_key: The key for the cluster column in `adata.obs`.
    :type cluster_key: str, optional
    :param output_cell_type_obs_column: The name of the observation column to store the predicted cell types.
    :type output_cell_type_obs_column: str, optional
    :param output_cell_type_confidence_score_obs_column: The name of the observation column to store the
      confidence scores.
    :type output_cell_type_confidence_score_obs_column: str, optional
    :param ignore_set: A set of cell types to ignore during the majority vote.
    :type ignore_set: set, optional
    :return: A dictionary containing detailed information about the majority vote for each cluster.
    :rtype: dict
    """
    if len(adata) != len(cas_cell_type_summary_statistics_response):
        raise ValueError("Lengths of `cas_query_res` and `adata` should be the same")
    if cluster_key not in adata.obs:
        raise ValueError("`cluster_key` does not correspond to a column in `adata.obs`")
    if ignore_set is None:
        ignore_set = set()
    output_cell_type_list = [None] * adata.shape[0]
    output_cell_type_confidence_score_list = [None] * adata.shape[0]
    cluster_detailed_info_dict = dict()
    for cluster_id in adata.obs[cluster_key].values.categories:
        cluster_cell_indices = np.where(adata.obs[cluster_key] == cluster_id)[0]
        if len(cluster_cell_indices) == 0:
            raise ValueError(f"No cells found belonging to cluster_id {cluster_id}")
        cluster_query = defaultdict(int)
        for cell_index in cluster_cell_indices:
            single_cell_query = cas_cell_type_summary_statistics_response[cell_index]
            for match in single_cell_query["matches"]:
                if match["cell_type"] in ignore_set:
                    continue
                cluster_query[match["cell_type"]] += match["cell_count"]
        total_cell_count = sum(cluster_query.values())
        sorted_hits_and_freqs = sorted(
            [(cell_type, count / total_cell_count) for cell_type, count in cluster_query.items()],
            key=itemgetter(1),
            reverse=True,
        )
        best_cell_type = sorted_hits_and_freqs[0][0]
        best_cell_type_confidence_score = sorted_hits_and_freqs[0][1]
        cluster_detailed_info_dict[cluster_id] = sorted_hits_and_freqs
        for cell_index in cluster_cell_indices:
            output_cell_type_list[cell_index] = best_cell_type
            output_cell_type_confidence_score_list[cell_index] = best_cell_type_confidence_score
    adata.obs[output_cell_type_obs_column] = output_cell_type_list
    adata.obs[output_cell_type_confidence_score_obs_column] = output_cell_type_confidence_score_list

    return cluster_detailed_info_dict


def _get_weights_via_fuzzy_simplicial_sets(
    i_obs: int,
    adata: AnnData,
    n_neighbors: int,
    self_connectivity: float,
    connectivities_key: str,
) -> tuple:
    assert connectivities_key in adata.obsp, "Please run `scanpy.pp.neighbors` first"
    connectivity_values = adata.obsp[connectivities_key][i_obs].data
    connectivity_indices = adata.obsp[connectivities_key][i_obs].indices

    # append self
    connectivity_values = np.append(connectivity_values, self_connectivity)
    connectivity_indices = np.append(connectivity_indices, i_obs)

    # convert to weights
    _order = np.argsort(connectivity_values)[::-1][:n_neighbors]
    sorted_connectivity_values = connectivity_values[_order]
    sorted_connectivity_indices = connectivity_indices[_order]
    weights = sorted_connectivity_values / np.sum(sorted_connectivity_values)
    return sorted_connectivity_indices, weights


def _get_cell_type_probs(single_cell_query: t.Dict, all_cell_types: t.List, cell_type_to_idx_map: t.Dict) -> np.ndarray:
    probs = np.zeros((len(all_cell_types),))
    for match in single_cell_query["matches"]:
        probs[cell_type_to_idx_map[match["cell_type"]]] += match["cell_count"]
    return probs / np.sum(probs)


def reduce_cas_cell_type_summary_statistics_response_by_wnn(
    adata: AnnData,
    cas_cell_type_summary_statistics_response: dict,
    n_neighbors: int = 10,
    wnn_strategy: str = "connectivities",
    connectivities_key: str = "connectivities",
    self_connectivity: float = 1.0,
    min_n_cells_per_type: int = 10,
    output_unreliable_type: str = "Unknown or Unconfident",
    output_cell_type_obs_column: str = "cas_cell_type",
    output_cell_type_confidence_score_obs_column: str = "cas_cell_type_confidence_score",
):
    """
    Reduce cell type summary statistics response by Weighted Nearest Neighbors (WNN).

    Args:
        adata (AnnData): Annotated data object.
        cas_cell_type_summary_statistics_response (dict): Cell type summary statistics response from CAS.
        n_neighbors (int, optional): Number of neighbors to consider. Defaults to 10.
        wnn_strategy (str, optional): WNN strategy. Defaults to "connectivities".
        connectivities_key (str, optional): Key for connectivities in `adata.uns['neighbors']`.
          Defaults to "connectivities".
        self_connectivity (float, optional): Self-connectivity value. Defaults to 1.0.
        min_n_cells_per_type (int, optional): Minimum number of cells per cell type.
          Defaults to 10.
        output_unreliable_type (str, optional): Output value for unreliable cell types.
          Defaults to "Unknown or Unconfident".
        output_cell_type_obs_column (str, optional): Output column name for cell type. Defaults to "cas_cell_type".
        output_cell_type_confidence_score_obs_column (str, optional): Output column name for cell type confidence score. Defaults to "cas_cell_type_confidence_score".

    Returns:
        tuple: A tuple containing the cell type to index mapping and a list of cell type probability arrays.
    """

    if wnn_strategy not in ALLOWED_WNN_STRATEGIES:
        raise ValueError(
            f"`wnn_strategy` should be one of {', '.join(ALLOWED_WNN_STRATEGIES)}, got {wnn_strategy} instead"
        )
    if n_neighbors > adata.uns["neighbors"]["params"]["n_neighbors"]:
        raise ValueError(
            "`n_neighbors` should be less than or equal to `adata.uns['neighbors']['params']['n_neighbors']`"
        )
    if len(adata) != len(cas_cell_type_summary_statistics_response):
        raise ValueError("Lengths of `cas_query_res` and `adata` should be the same")

    all_cell_types = []
    for single_cell_query in cas_cell_type_summary_statistics_response:
        for match in single_cell_query["matches"]:
            all_cell_types.append(match["cell_type"])
    all_cell_types = list(set(all_cell_types))
    cell_type_to_idx_map = {cell_type: idx for idx, cell_type in enumerate(all_cell_types)}

    cell_type_probs_list = []
    majority_vote_cell_type_list = []
    majority_vote_confidence_score_list = []

    n_cells = len(cas_cell_type_summary_statistics_response)
    for i in tqdm(range(n_cells)):
        neighbor_indices, neighbor_weights = _get_weights_via_fuzzy_simplicial_sets(
            i,
            adata,
            n_neighbors,
            self_connectivity,
            connectivities_key=connectivities_key,
        )
        if not np.isclose(np.sum(neighbor_weights), 1.0):
            raise ValueError("Sum of Neighbor weights is not close to 1")

        cell_type_probs = np.zeros((len(all_cell_types),))
        for j, weight in zip(neighbor_indices, neighbor_weights):
            cell_type_probs += weight * _get_cell_type_probs(
                cas_cell_type_summary_statistics_response[j], all_cell_types, cell_type_to_idx_map
            )

        best_cell_type_idx = np.argmax(cell_type_probs)
        best_cell_type = all_cell_types[best_cell_type_idx]
        best_cell_type_prob = cell_type_probs[best_cell_type_idx]

        cell_type_probs_list.append(cell_type_probs)
        majority_vote_cell_type_list.append(best_cell_type)
        majority_vote_confidence_score_list.append(best_cell_type_prob)

    all_cell_types = set(majority_vote_cell_type_list)
    update_map = dict()
    for cell_type in all_cell_types:
        n_counts = sum([_cell_type == cell_type for _cell_type in majority_vote_cell_type_list])
        if n_counts < min_n_cells_per_type:
            update_map[cell_type] = output_unreliable_type
        else:
            update_map[cell_type] = cell_type
    majority_vote_cell_type_list = list(map(update_map.get, majority_vote_cell_type_list))

    adata.obs[output_cell_type_obs_column] = majority_vote_cell_type_list
    adata.obs[output_cell_type_confidence_score_obs_column] = majority_vote_confidence_score_list

    return cell_type_to_idx_map, cell_type_probs_list


def hex_to_rgb(value: str) -> np.ndarray:
    """
    Convert a hexadecimal color value to an RGB array.

    :param value: The hexadecimal color value to convert.
    :type value: str
    :return: The RGB array representing the color.
    :rtype: np.ndarray
    """
    value = value.lstrip("#")
    lv = len(value)
    return np.asarray(tuple(float(int(value[i : i + lv // 3], 16)) for i in range(0, lv, lv // 3)))


def rgb_to_tuple(value: np.ndarray) -> tuple:
    """
    Convert an RGB value represented as a NumPy array to a tuple of integers.

    :param value: The RGB value as a NumPy array.
    :type value: np.ndarray
    :return: The RGB value as a tuple of integers.
    :rtype: tuple
    """
    return int(value[0]), int(value[1]), int(value[2])


def rgb_to_hex(r: int, g: int, b: int) -> str:
    """
    Convert RGB values to a hexadecimal color code.

    :param r: The red component of the RGB color (0-255).
    :param g: The green component of the RGB color (0-255).
    :param b: The blue component of the RGB color (0-255).
    :return: The hexadecimal color code representing the RGB color.
    """
    return "#%02x%02x%02x" % (r, g, b)


def get_interpolated_cell_type_colors(
    adata: AnnData,
    cell_type_to_idx_map: t.Dict[str, int],
    cell_type_probs_list: t.List[np.ndarray],
    na_cell_type_key: str = "Unknown or Unconfident",
) -> np.ndarray:
    """Get interpolated cell type colors based on the cell type probabilities.

    This function takes the cell type probabilities and maps them to the cell type colors. The cell type colors are
    interpolated based on the probabilities.

    :param adata: An Annotated Data object containing the query result.
    :type adata: anndata.AnnData
    :param cell_type_to_idx_map: A dictionary mapping cell types to indices.
    :type cell_type_to_idx_map: dict
    :param cell_type_probs_list: A list of cell type probabilities.
    :type cell_type_probs_list: list
    :param na_cell_type_key: The key for the unknown cell type.
    :type na_cell_type_key: str, optional
    :return: The interpolated cell type colors.
    :rtype: np.ndarray
    """
    idx_to_cell_type_map: t.Dict[int, str] = {idx: cell_type for cell_type, idx in cell_type_to_idx_map.items()}
    all_cell_types: t.List[str] = list(map(idx_to_cell_type_map.get, range(len(cell_type_to_idx_map))))
    plot_cell_type_colors: np.ndarray = adata.uns["cas_cell_type_colors"]
    plot_cell_types: t.List[str] = list(adata.obs["cas_cell_type"].values.categories)
    plot_cell_types_to_idx_map: t.Dict[str, int] = {cell_type: idx for idx, cell_type in enumerate(plot_cell_types)}

    cell_type_probs_nk: np.ndarray = np.asarray(cell_type_probs_list)
    cell_type_map_kq: np.ndarray = np.zeros((len(all_cell_types), len(plot_cell_types)))
    na_index: int = plot_cell_types.index(na_cell_type_key)
    plot_cell_type_colors[na_index] = rgb_to_hex(r=255, g=255, b=255)
    for k, input_cell_type in enumerate(all_cell_types):
        if input_cell_type in plot_cell_types_to_idx_map:
            q: int = plot_cell_types_to_idx_map[input_cell_type]
        else:
            q: int = na_index
        cell_type_map_kq[k, q] += 1

    plot_colors_q3: np.ndarray = np.asarray(list(hex_to_rgb(hex_color) for hex_color in plot_cell_type_colors))
    cell_colors_n3: np.ndarray = (cell_type_probs_nk @ cell_type_map_kq @ plot_colors_q3) / 255

    return np.clip(cell_colors_n3, 0.0, 1.0)
